"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
from typing import Literal, Union

Font = Literal["1943____","1row","3-d","3d-ascii","3d_diagonal","3x5","4max","4x4_offr","5lineoblique","5x7","5x8","64f1____","6x10","6x9","acrobatic","advenger","alligator","alligator2","alpha","alphabet","amc_3_line","amc_3_liv1","amc_aaa01","amc_neko","amc_razor","amc_razor2","amc_slash","amc_slider","amc_thin","amc_tubes","amc_untitled","ansi_regular","ansi_shadow","aquaplan","arrows","ascii_new_roman","ascii___","asc_____","assalt_m","asslt__m","atc_gran","atc_____","avatar","a_zooloo","b1ff","banner","banner3-D","banner3","banner4","barbwire","basic","battlesh","battle_s","baz__bil","bear","beer_pub","bell","benjamin","big","bigchief","bigfig","big_money-ne","big_money-nw","big_money-se","big_money-sw","binary","block","blocks","blocky","bloody","bolger","braced","bright","brite","briteb","britebi","britei","broadway","broadway_kb","bubble","bubble_b","bubble__","bulbhead","b_m__200","c1______","c2______","calgphy2","caligraphy","calvin_s","cards","catwalk","caus_in_","char1___","char2___","char3___","char4___","charact1","charact2","charact3","charact4","charact5","charact6","characte","charset_","chartr","chartri","chiseled","chunky","clb6x10","clb8x10","clb8x8","cli8x8","clr4x6","clr5x10","clr5x6","clr5x8","clr6x10","clr6x6","clr6x8","clr7x10","clr7x8","clr8x10","clr8x8","coil_cop","coinstak","cola","colossal","computer","com_sen_","contessa","contrast","convoy__","cosmic","cosmike","cour","courb","courbi","couri","crawford","crawford2","crazy","cricket","cursive","cyberlarge","cybermedium","cybersmall","cygnet","c_ascii_","c_consen","danc4","dancing_font","dcs_bfmo","decimal","deep_str","defleppard","def_leppard","delta_corps_priest_1","demo_1__","demo_2__","demo_m__","devilish","diamond","diet_cola","digital","doh","doom","dos_rebel","dotmatrix","double","double_shorts","drpepper","druid___","dwhistled","d_dragon","ebbs_1__","ebbs_2__","eca_____","eftichess","eftifont","eftipiti","eftirobot","eftitalic","eftiwall","eftiwater","efti_robot","electronic","elite","epic","etcrvs__","e__fist_","f15_____","faces_of","fairligh","fair_mea","fantasy_","fbr12___","fbr1____","fbr2____","fbr_stri","fbr_tilt","fender","filter","finalass","fireing_","fire_font-k","fire_font-s","flipped","flower_power","flyn_sh","fourtops","fp1_____","fp2_____","fraktur","funky_dr","fun_face","fun_faces","future_1","future_2","future_3","future_4","future_5","future_6","future_7","future_8","fuzzy","gauntlet","georgi16","georgia11","ghost","ghost_bo","ghoulish","glenyn","goofy","gothic","gothic__","graceful","gradient","graffiti","grand_pr","greek","green_be","hades___","heart_left","heart_right","heavy_me","helv","helvb","helvbi","helvi","henry_3d","heroboti","hex","hieroglyphs","high_noo","hills___","hollywood","home_pak","horizontal_left","horizontal_right","house_of","hypa_bal","hyper___","icl-1900","impossible","inc_raw_","invita","isometric1","isometric2","isometric3","isometric4","italic","italics_","ivrit","jacky","jazmine","jerusalem","joust___","js_block_letters","js_bracket_letters","js_capital_curves","js_cursive","js_stick_letters","katakana","kban","keyboard","kgames_i","kik_star","knob","konto","konto_slant","krak_out","larry3d","lazy_jon","lcd","lean","letters","letterw3","letter_w","lexible_","lil_devil","line_blocks","linux","lockergnome","madrid","mad_nurs","magic_ma","marquee","master_o","maxfour","mayhem_d","mcg_____","merlin1","merlin2","mig_ally","mike","mini","mirror","mnemonic","modern__","modular","morse","morse2","moscow","mshebrew210","muzzle","nancyj-fancy","nancyj-improved","nancyj-underlined","nancyj","new_asci","nfi1____","nipples","notie_ca","npn_____","nscript","ntgreek","nvscript","o8","octal","odel_lak","ogre","ok_beer_","old_banner","os2","outrun__","pacos_pe","panther_","patorjk's_cheese","patorjk-hex","pawn_ins","pawp","peaks","pebbles","pepper","phonix__","platoon2","platoon_","pod_____","poison","puffy","puzzle","pyramid","p_skateb","p_s_h_m_","r2-d2___","radical_","rad_phan","rad_____","rainbow_","rally_s2","rally_sp","rammstein","rampage_","rastan__","raw_recu","rci_____","rectangles","red_phoenix","relief","relief2","rev","ripper!_","road_rai","rockbox_","rok_____","roman","roman___","rot13","rotated","rounded","rowancap","rozzo","runic","runyc","sans","sansb","sansbi","sansi","santa_clara","sblood","sbook","sbookb","sbookbi","sbooki","script","script__","serifcap","shadow","shimrod","short","skateord","skateroc","skate_ro","sketch_s","slant","slant_relief","slide","slscript","sl_script","small","small_caps","small_poison","small_shadow","small_slant","smisome1","smkeyboard","smscript","smshadow","smslant","smtengwar","sm______","soft","space_op","spc_demo","speed","spliff","stacey","stampate","stampatello","standard","starwars","star_strips","star_war","stealth_","stellar","stencil1","stencil2","stforek","stick_letters","stop","straight","street_s","stronger_than_all","sub-zero","subteran","super_te","swamp_land","swan","sweet","tanja","tav1____","taxi____","tec1____","tecrvs__","tec_7000","tengwar","term","test1","the_edge","thick","thin","this","thorned","threepoint","ticks","ticksslant","tiles","times","timesofl","tinker-toy","ti_pan__","tomahawk","tombstone","top_duck","train","trashman","trek","triad_st","ts1_____","tsalagi","tsm_____","tsn_base","tty","ttyb","tubular","twin_cob","twisted","twopoint","type_set","t__of_ap","ucf_fan_","ugalympi","unarmed_","univers","usaflag","usa_pq__","usa_____","utopia","utopiab","utopiabi","utopiai","varsity","vortron_","war_of_w","wavy","weird","wet_letter","whimsy","wow","xbrite","xbriteb","xbritebi","xbritei","xchartr","xchartri","xcour","xcourb","xcourbi","xcouri","xhelv","xhelvb","xhelvbi","xhelvi","xsans","xsansb","xsansbi","xsansi","xsbook","xsbookb","xsbookbi","xsbooki","xtimes","xtty","xttyb","yie-ar__","yie_ar_k","z-pilot_","zig_zag_","zone7___"]
Image = Literal["bruh","bruh_empty","bruh_computer","computer", "hey","twopoint","christmas"]
EffectType = Literal["static","offset","noise","stars","plasma","gol","rain","matrix","drawlines","snow","twinkle","audio","chat"]
valid_effect_types = {"static","offset","noise","stars","plasma","gol","rain","matrix","drawlines","snow","twinkle","audio","chat"}
PanRendererDirection = Literal["horizontal", "vertical"]
valid_pan_renderer_directions = {"horizontal", "vertical"}



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import random

from bruhcolor import bruhcolored
from ..bruhutil import FLAKE_WEIGHT_CHARS, SNOWFLAKE_TYPES, SNOWFLAKE_COLORS, Buffer
from .base_effect import BaseEffect


class SnowEffect(BaseEffect):
    def __init__(
        self,
        buffer: Buffer,
        background,
        img_start_x=None,
        img_start_y=None,
        img_width=None,
        img_height=None,
        collision=False,
        show_info=False,
    ):
        super(SnowEffect, self).__init__(buffer, background)
        self.image_present = (True if img_start_x and img_start_y and img_width and img_height else False)
        self.collision = collision
        self.total_ground_flakes = 0
        self.show_info = show_info
        self.flakes = []
        self.ground_flakes = [[0 for _ in range(buffer.width())] for _ in range(buffer.height())]
        self.image_flakes = [None for _ in range(self.buffer.width())]
        self.smart_transparent = False

    def update_collision(
        self,
        img_start_x,
        img_start_y,
        img_width,
        img_height,
        collision,
        smart_transparent,
        image_buffer=None,
    ):
        """
        Function to set whether or not to visually see the snow collide with the ground
        or images if they are present
        :param img_start_x: where the image starts on the screen
        :param img_start_y: where the image starts on the screen
        :param img_width:   the width of the image
        :param img_height:  the height of the image
        :param collision:   update collision variable
        """
        self.image_present = (
            True if img_start_x and img_start_y and img_width and img_height else False
        )
        self.collision = collision
        if self.image_present:
            self.img_start_x = img_start_x
            self.img_start_y = img_start_y
            self.img_height = img_height
            self.img_width = img_width
            self.img_end_y = img_start_y + img_height
            self.image_buffer = image_buffer
            self.image_x_boundaries = (img_start_x, img_start_x + img_width)
            self.image_y_boundaries = (img_start_y, img_start_y + img_height)
        else:
            self.image_buffer = None

    def set_show_info(self, show_info: bool):
        self.show_info = show_info

    def generate_snowflake(self, x: int):
        snowflake_type = random.choice(list(SNOWFLAKE_TYPES.keys()))
        snowflake_char = bruhcolored(snowflake_type, SNOWFLAKE_COLORS[snowflake_type])
        self.flakes.append({"x": x, "y": 0, "type": snowflake_type, "colored": snowflake_char, "fall_delay": 0})

    def can_stack(self, x, y):
        if y >= self.buffer.height():
            return False
        return self.ground_flakes[y][x] >= 18

    def is_colliding(self, x: int, y: int):
        if self.image_present:
            if self.image_x_boundaries[0] < x < self.image_x_boundaries[1] and self.image_y_boundaries[0] < y < self.image_y_boundaries[1]:
                return True
        return False

    def handle_snowflake_landing(self, x: int, y: int):
        self.ground_flakes[y][x] += 1
        weight = self.ground_flakes[y][x]
        for w, char in sorted(FLAKE_WEIGHT_CHARS.items(), reverse=True):
            if weight >= w:
                self.buffer.put_char(x, y, char)
                break
        if weight > 18 and y > 0:
            self.ground_flakes[y - 1][x] += 1

    def add_info(self):
        self.buffer.put_at(0, 0, f"Total Snow Flakes: {len(self.flakes)}")
        self.buffer.put_at(0, 1, f"Total Flakes on Ground: {sum([sum([1 for v in row if v > 0]) for row in self.ground_flakes])}")

    def render_frame(self, frame_number):
        """
        Function to render the next frame of the snow effect into it's buffer.
        :param frame_number: The current frame number (used to determine the animation state).
        """

        # generate a new row of snowflakes
        for idx in range(self.buffer.width()):
            if random.random() < 0.01:
                self.generate_snowflake(idx)
    
        # update the positions of all flakes
        new_flakes = []
        for snowflake in self.flakes:
            x, y, flake_type = snowflake["x"], snowflake["y"], snowflake["type"]
            speed = SNOWFLAKE_TYPES[flake_type]["speed"]

            # clear out the old position of the snowflake
            self.buffer.put_char(x, y, " ")

            if snowflake["fall_delay"] < speed:
                snowflake["fall_delay"] += 1
                new_flakes.append(snowflake)
                continue
            else:
                snowflake["fall_delay"] = 0

            if y + 1 >= self.buffer.height() or self.can_stack(x, y + 1):
                self.handle_snowflake_landing(x, y)
            else:
                dx = random.choice([-1, 0, 1])
                dy = 1

                new_x = max(0, min(self.buffer.width() - 1, x + dx))
                new_y = y + dy
                if new_y < self.buffer.height() and not self.is_colliding(new_x, new_y) and 0 <= new_x < self.buffer.width():
                    snowflake["x"], snowflake["y"] = new_x, new_y
                    new_flakes.append(snowflake)
                else:
                    self.handle_snowflake_landing(x, y)

        self.flakes = new_flakes

        # place the updates into the buffer
        for snowflake in self.flakes:
            x, y, flake_type, colored = snowflake["x"], snowflake["y"], snowflake["type"], snowflake["colored"]
            self.buffer.put_char(x, y, colored.colored)
        
        if self.show_info:
            self.add_info()



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import random
from typing import List
from .base_renderer import BaseRenderer
from ..bruhutil.bruhtypes import EffectType


class FocusRenderer(BaseRenderer):
    """
    A Renderer that takes an image and randomly spreads the characters around the screen.
    The characters are then pulled to the middle of the screen
    """

    def __init__(
        self,
        screen,
        img: List[str],
        frames: int = 100,
        time: float = 0.1,
        effect_type: EffectType = "static",
        background: str = " ",
        transparent: bool = False,
        collision: bool = False,
        start_frame: int = 0,
        reverse: bool = False,
        start_reverse: int = None,
        loop: bool = True
    ):
        super(FocusRenderer, self).__init__(
            screen, frames, time, effect_type, background, transparent, collision
        )
        self.background = background if background else " "
        self.transparent = transparent if transparent else False
        self.img = img
        self.start_frame = start_frame
        self.reverse = reverse
        self.start_reverse = start_reverse
        if self.start_reverse:
            self.frame_gap = start_reverse - start_frame
        self.loop = True if loop == True and reverse == True else False
        self.loops = 1

        if self.reverse and self.start_reverse == None:
            raise Exception(
                "if reverse is enabled, and start_reverse frame must be provided"
            )

        if self.reverse and start_reverse < self.start_frame:
            raise Exception(
                f"the frame to start the reverse can not be less than the start frame\n\tstart_frame: {self.start_frame}, start_reverse: {self.start_reverse}"
            )

        if self.img:
            self._set_img_attributes()

    def _set_img_attributes(self):
        """
        Set the attributes of the img
        """
        self.img_height = len(self.img)
        self.img_width = len(self.img[0])
        self.img_y_start = (self.height - len(self.img)) // 2
        self.img_x_start = (self.width - len(self.img[0])) // 2
        self.current_img_x = self.img_x_start
        self.current_img_y = self.img_y_start
        self.start_board = [
            [
                [
                    random.randint(0, self.width - 1),
                    random.randint(0, self.height - 1),
                    self.img[y][x],
                    (x, y),
                ]
                for x in range(self.img_width)
            ]
            for y in range(self.img_height)
        ]
        self.current_board = [
            [
                [
                    self.start_board[y][x][0],
                    self.start_board[y][x][1],
                    self.img[y][x],
                    (x, y),
                ]
                for x in range(self.img_width)
            ]
            for y in range(self.img_height)
        ]
        self.end_board = [
            [
                [self.img_x_start + x, self.img_y_start + y, self.img[y][x], (x, y)]
                for x in range(self.img_width)
            ]
            for y in range(self.img_height)
        ]
        self.direction_board = [
            [
                [
                    -1
                    if (self.end_board[y][x][0] - self.current_board[y][x][0]) < 0
                    else 1,
                    -1
                    if (self.end_board[y][x][1] - self.current_board[y][x][1]) < 0
                    else 1,
                ]
                for x in range(self.img_width)
            ]
            for y in range(self.img_height)
        ]

    def update_reverse(self, reverse, start_reverse):
        """
        Function to update whether or not to reverse the Focus
        :param reverse: True / False
        """
        self.reverse = reverse
        self.start_reverse = start_reverse
        if start_reverse < self.start_frame:
            raise Exception(
                f"the frame to start the reverse can not be less than the start frame\n\tstart_frame: {self.start_frame}, start_reverse: {self.start_reverse}"
            )

    def update_start_frame(self, frame_number):
        """
        Updates the frame at which the Focus Effect should start
        :param frame_number: Frame to start
        """
        self.start_frame = frame_number

    def solved(self, end_state):
        """
        Function that determines if the image has been moved back to its
        original shape
        """
        b1 = self.current_board
        b2 = self.end_board
        b3 = self.start_board
        if end_state == "end":
            for y in range(len(b1)):
                for x in range(len(b1[y])):
                    if b1[y][x][0] == b2[y][x][0] and b1[y][x][1] == b2[y][x][1]:
                        pass
                    else:
                        return False
            return True
        elif end_state == "start":
            for y in range(len(b1)):
                for x in range(len(b1[y])):
                    if b1[y][x][0] == b3[y][x][0] and b1[y][x][1] == b3[y][x][1]:
                        pass
                    else:
                        return False
            return True
        else:
            raise Exception(f"unkown solved board state for FocusRenderer: {end_state}")

    def render_img_frame(self, frame_number):
        """
        Renders the next image frame into the image buffer
        """        
        if self.reverse and frame_number >= self.start_reverse:
            if not self.solved(end_state="start"):
                for y in range(len(self.current_board)):
                    for x in range(len(self.current_board[y])):
                        x_check, y_check = False, False
                        # MOVE X IF NEEDED
                        if self.current_board[y][x][0] != self.start_board[y][x][0]:
                            self.current_board[y][x][0] -= self.direction_board[y][x][0]
                        else:
                            x_check = True
                        # MOVE Y IF NEEDED
                        if self.current_board[y][x][1] != self.start_board[y][x][1]:
                            self.current_board[y][x][1] -= self.direction_board[y][x][1]
                        else:
                            y_check = True

                        if x_check and y_check:
                            self.current_board[y][x][2] = None
                self.image_buffer.clear_buffer(val=None)
                for row in self.current_board:
                    for value in row:
                        self.image_buffer.put_char(
                            value[0], value[1], value[2], transparent=self.transparent
                        )
            else:
                if self.loop:
                    self.loops += 1
                    self.start_frame = self.start_reverse + self.frame_gap
                    self.start_reverse = self.start_frame + self.frame_gap
                    self._set_img_attributes()
                self.image_buffer.clear_buffer(val=None)
        elif frame_number >= self.start_frame:
            if not self.solved(end_state="end"):
                for y in range(len(self.current_board)):
                    for x in range(len(self.current_board[y])):
                        # MOVE X IF NEEDED
                        if self.current_board[y][x][0] != self.end_board[y][x][0]:
                            self.current_board[y][x][0] += self.direction_board[y][x][0]

                        # MOVE Y IF NEEDED
                        if self.current_board[y][x][1] != self.end_board[y][x][1]:
                            self.current_board[y][x][1] += self.direction_board[y][x][1]
                self.image_buffer.clear_buffer(val=None)
                for row in self.current_board:
                    for value in row:
                        self.image_buffer.put_char(
                            value[0], value[1], value[2], transparent=self.transparent
                        )




"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
from typing import Literal, TypeVar, Dict, Generic
from abc import abstractmethod
from ..bruheffect import *
from ..bruhutil.utils import sleep
from ..bruhutil.bruhffer import Buffer
from ..bruhutil.bruhscreen import Screen
from ..bruhutil.bruherrors import InvalidEffectTypeError
from ..bruhutil.bruhtypes import EffectType, valid_effect_types


HORIZONTAL = "h"
VERTICAL = "v"
INF = float("inf")

class BaseRenderer:
    """
    Base class for rendering effects on the screen.
    """

    def __init__(
        self,
        screen: Screen,
        frames: int = 100,
        frame_time: float = 0.1,
        effect_type: EffectType = "static",
        background: str = " ",
        transparent: bool = False,
        collision: bool = False,
    ):
        self.validate_effect_type(effect_type)
        
        self.screen = screen
        self.frames = frames
        self.frame_time = frame_time
        self.effect_type = effect_type
        self.transparent = transparent
        self.background = background
        self.height = screen.height
        self.width = screen.width
        self.smart_transparent = False
        self.collision = collision

        self.effect = self.create_effect(effect_type)
        self.image_buffer = self.create_buffer().clear_buffer(val=None)
        self.back_buffer = self.create_buffer()
        self.front_buffer = self.create_buffer()

        self.exit_messages = {
            "msg1": " Frames Are Done ",
            "msg2": "   Press Enter   ",
            "centered": True,
            "wipe": False,
            "x_loc": 0,
            "y_loc": 1,
        }

    def validate_effect_type(self, effect_type: EffectType):
        """Validate the effect type against available effects."""
        if effect_type not in valid_effect_types:
            raise InvalidEffectTypeError(f"'{effect_type}' is not a valid effect. Please choose from {valid_effect_types}")

    def create_effect(self, effect_type: EffectType):
        """Create an effect instance based on the specified type."""
        if effect_type == "static":
            return StaticEffect(self.create_buffer(), self.background)
        elif effect_type == "offset":
            return OffsetEffect(self.create_buffer(), self.background)
        elif effect_type == "noise":
            return NoiseEffect(self.create_buffer(), self.background)
        elif effect_type == "stars":
            return StarEffect(self.create_buffer(), self.background)
        elif effect_type == "plasma":
            return PlasmaEffect(self.create_buffer(), self.background)
        elif effect_type == "gol":
            return GameOfLifeEffect(self.create_buffer(), self.background)
        elif effect_type == "rain":
            return RainEffect(self.create_buffer(), self.background)
        elif effect_type == "matrix":
            return MatrixEffect(self.create_buffer(), self.background)
        elif effect_type == "drawlines":
            return DrawLinesEffect(self.create_buffer(), self.background)
        elif effect_type == "snow":
            return SnowEffect(self.create_buffer(), self.background)
        elif effect_type == "twinkle":
            return TwinkleEffect(self.create_buffer(), self.background)
        elif effect_type == "audio":
            return AudioEffect(self.create_buffer(), self.background)
        elif effect_type == "chat":
            return ChatbotEffect(self.screen, self.create_buffer(), self.create_buffer(), self.background)

    def create_buffer(self) -> Buffer:
        return Buffer(height=self.screen.height, width=self.screen.width)

    def update_collision(self, collision):
        """Update collision for specific effects."""
        try:
            self.collision = collision
            self.effect.update_collision(
                self.current_img_x,
                self.current_img_y,
                self.img_width,
                self.img_height,
                collision,
                self.smart_transparent,
                self.image_buffer,
            )
        except Exception as e:
            print(f"base_renderer: update_collision: warning: {e}")
            self.effect.update_collision(None, None, None, None, collision, None)

    def update_smart_transparent(self, smart_transparent: bool):
        """Enable/Disable the smart transparency effect."""
        self.smart_transparent = smart_transparent
        self.effect.smart_transparent = smart_transparent

    def push_front_to_screen(self):
        """Pushes changes from front_buffer to the screen."""
        for y, x, val in self.front_buffer.get_buffer_changes(self.back_buffer):
            self.screen.print_at(val, x, y, 1)

    def render_exit(self):
        """Renders the exit prompt to the screen."""
        if self.exit_messages['wipe']:
            self.back_buffer.clear_buffer()
        if self.exit_messages['centered']:
            self.back_buffer.put_at_center(self.height // 2 - 1, self.exit_messages['msg1'])
            self.back_buffer.put_at_center(self.height // 2, self.exit_messages['msg2'])
        else:
            self.back_buffer.put_at(self.exit_messages['x_loc'], self.exit_messages['y_loc'] - 1, self.exit_messages['msg1'], transparent=False)
            self.back_buffer.put_at(self.exit_messages['x_loc'], self.exit_messages['y_loc'], self.exit_messages['msg2'], transparent=False)

    def run(self, end_message=True):
        """Main loop for rendering frames."""
        try:
            if self.frames != INF:
                for frame in range(self.frames):
                    if self.screen.has_resized():
                        raise Exception("The screen was resized.")
                    sleep(self.frame_time)
                    self.render_img_frame(frame)
                    self.effect.render_frame(frame)
                    self.back_buffer.sync_with(self.effect.buffer)
                    self.back_buffer.sync_over_top(self.image_buffer)
                    self.push_front_to_screen()
                    self.front_buffer.sync_with(self.back_buffer)
            else:
                frame = 0
                while True:
                    if self.screen.has_resized():
                        raise Exception("The screen was resized.")
                    sleep(self.frame_time)
                    self.render_img_frame(frame)
                    self.effect.render_frame(frame)
                    self.back_buffer.sync_with(self.effect.buffer)
                    self.back_buffer.sync_over_top(self.image_buffer)
                    self.push_front_to_screen()
                    self.front_buffer.sync_with(self.back_buffer)
                    frame += 1

            if end_message:
                self.render_exit()
                self.push_front_to_screen()
                if sys.platform == 'win32':
                    input()
        except KeyboardInterrupt:
            if end_message:
                self.render_exit()
                self.push_front_to_screen()
                if sys.platform == 'win32':
                    input()

    def update_exit_stats(
        self, msg1=None, msg2=None, wipe=None, x_loc=None, y_loc=None, centered=False
    ):
        """Set the exit messages for when the animation finishes."""
        if msg1:
            self.exit_messages['msg1'] = msg1.replace("\n", "")
        if msg2:
            self.exit_messages['msg2'] = msg2.replace("\n", "")
        if wipe is not None:
            self.exit_messages['wipe'] = wipe
        if x_loc is not None and y_loc is not None:
            self.exit_messages['x_loc'], self.exit_messages['y_loc'] = x_loc, y_loc
        self.exit_messages['centered'] = centered

    @abstractmethod
    def render_frame(self):
        """To be defined by each renderer."""



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import random

from bruhcolor import bruhcolored
from ..bruhutil import LIFE_COLORS, LIFE_SCALES
from .base_effect import BaseEffect


class GameOfLifeEffect(BaseEffect):
    """
    Effect ot simulate Conway's Game of Life
    """

    def __init__(
        self,
        buffer,
        background,
        decay=False,
        color=False,
        color_type=None,
        scale="random",
    ):
        super(GameOfLifeEffect, self).__init__(buffer, background)
        self.decay = decay
        self.scale = scale
        self.color = color
        self.color_type = color_type
        self._set_attributes()
        self.direcitons = [
            [1, 0],
            [0, 1],
            [-1, 0],
            [0, -1],
            [1, 1],
            [1, -1],
            [-1, 1],
            [-1, -1],
        ]
        self.rules = {"life": [2, 3], "death": [3, 3]}
        self.board = [
            [" " for _ in range(self.buffer.width())]
            for __ in range(self.buffer.height())
        ]

    def _set_attributes(self):
        """
        Function to set the attributes of the effect
        """
        self.grey_scale = (
            LIFE_SCALES[random.choice(list(LIFE_SCALES.keys()))]
            if self.decay and self.scale == "random"
            else LIFE_SCALES[self.scale] if self.decay else " o"
        )
        self.colors = [232, 231] if not self.decay else LIFE_COLORS[self.color_type]
        self.ALIVE = len(self.grey_scale) - 1
        self.DEAD = 0
        self.mappings = {i: self.grey_scale[i] for i in range(len(self.grey_scale))}

    def update_decay(self, decay, color_type="GREYSCALE", scale="random"):
        """
        Function to enable to decay and select the color map
        :param decay: True / False
        :param color_type: color map for the effect
        """
        self.decay = decay
        self.scale = scale
        if color_type:
            self.color_type = color_type
        self._set_attributes()

    def update_rules(self, life_rule, death_rule):
        self.rules["life"] = life_rule
        self.rules["death"] = death_rule

    def render_frame(self, frame_number):
        """
        Function to render the next frame of the GOL effect
        """
        if frame_number == 0:  # INITIALIZE THE GAME
            for y in range(self.buffer.height()):
                for x in range(self.buffer.width()):
                    if random.random() < 0.1:
                        self.buffer.put_char(
                            x,
                            y,
                            bruhcolored(
                                self.grey_scale[self.ALIVE],
                                color=self.colors[self.ALIVE],
                            ).colored,
                        )
                        self.board[y][x] = (
                            bruhcolored(
                                self.grey_scale[self.ALIVE],
                                color=self.colors[self.ALIVE],
                            ),
                            self.ALIVE,
                        )
                    else:
                        self.buffer.put_char(
                            x,
                            y,
                            bruhcolored(
                                self.grey_scale[self.DEAD], color=self.colors[self.DEAD]
                            ).colored,
                        )
                        self.board[y][x] = (
                            bruhcolored(
                                self.grey_scale[self.DEAD], color=self.colors[self.DEAD]
                            ),
                            0,
                        )
        else:  # RUN THE GAME
            all_neighbors = [
                [0 for _ in range(self.buffer.width())]
                for __ in range(self.buffer.height())
            ]
            for y in range(len(all_neighbors)):
                for x in range(len(all_neighbors[y])):
                    neighbors = 0
                    for direction in self.direcitons:
                        if (
                            0 <= y + direction[0] < self.buffer.height()
                            and 0 <= x + direction[1] < self.buffer.width()
                            and self.board[y + direction[0]][x + direction[1]][1]
                            == self.ALIVE
                        ):
                            neighbors += 1
                    all_neighbors[y][x] = neighbors
            for y in range(self.buffer.height()):
                for x in range(self.buffer.width()):
                    if self.board[y][x][1] == self.ALIVE:  # ALIVE
                        if (
                            self.rules["life"][0]
                            <= all_neighbors[y][x]
                            <= self.rules["life"][1]
                        ):  # STAY ALIVE
                            pass
                        else:  # MOVE TO THE FIRST DECAY STAGE
                            self.buffer.put_char(
                                x,
                                y,
                                bruhcolored(
                                    self.grey_scale[self.ALIVE - 1],
                                    color=self.colors[self.ALIVE - 1],
                                ).colored,
                            )
                            self.board[y][x] = (
                                bruhcolored(
                                    self.grey_scale[self.ALIVE - 1],
                                    color=self.colors[self.ALIVE - 1],
                                ),
                                self.ALIVE - 1,
                            )
                    else:  # DEAD
                        if (
                            self.rules["death"][0]
                            <= all_neighbors[y][x]
                            <= self.rules["death"][1]
                        ):  # COME BACK TO LIFE
                            self.buffer.put_char(
                                x,
                                y,
                                bruhcolored(
                                    self.grey_scale[self.ALIVE],
                                    color=self.colors[self.ALIVE],
                                ).colored,
                            )
                            self.board[y][x] = (
                                bruhcolored(
                                    self.grey_scale[self.ALIVE],
                                    color=self.colors[self.ALIVE],
                                ),
                                self.ALIVE,
                            )
                        else:  # MOVE TO THE NEXT STAGE --> IF AT 0 STAY AT 0 i.e. don't decrement
                            current_greyscale_position = self.board[y][x][1]
                            current_greyscale_position = (
                                current_greyscale_position - 1
                                if current_greyscale_position > 0
                                else 0
                            )
                            self.buffer.put_char(
                                x,
                                y,
                                bruhcolored(
                                    self.grey_scale[current_greyscale_position],
                                    color=self.colors[current_greyscale_position],
                                ).colored,
                            )
                            self.board[y][x] = (
                                bruhcolored(
                                    self.grey_scale[current_greyscale_position],
                                    color=self.colors[current_greyscale_position],
                                ),
                                current_greyscale_position,
                            )



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import math
import random
import requests
import openai
import string

from threading import Thread
from bruhcolor import bruhcolored
from ..bruhutil import Screen
from .base_effect import BaseEffect


class Key:
    def __init__(self, character, representation, value, x, y):
        self.x = x
        self.y = y
        self.character = character
        self.representation = representation
        self.value = value

    def __str__(self):
        return self.character

    def __repr__(self):
        return self.character


class GradientNoise:
    def __init__(self, x, y, length, char_halt=1, color_halt=1, gradient_length=1):
        self.x = x
        self.y = y
        self.__gradient_length = gradient_length
        # colors to use for gradient
        self.__gradient = [
            c
            for c in [
                232,
                232,
                232,
                232,
                233,
                233,
                233,
                233,
                234,
                234,
                234,
                234,
                235,
                235,
                235,
                235,
                236,
                236,
                236,
                236,
                237,
                238,
                239,
                240,
                241,
                242,
                243,
                244,
                245,
                246,
                247,
                248,
                249,
                250,
                251,
                252,
                253,
                254,
                255,
            ]
            for _ in range(self.__gradient_length)
        ]
        # delay to changing the chars in the noise
        self.__char_halt = char_halt
        self.__char_frame_number = 0
        # delay to change the gradient shift
        self.__color_halt = color_halt
        self.__color_frame_number = 0
        self.length = length
        self.done_generating = False

        self.string_chars = [" " for _ in range(self.length)]
        self.string_colors = [
            self.__gradient[i % len(self.__gradient)] for i in range(self.length)
        ]
        self.colored_chars = [
            bruhcolored(c, color=color)
            for c, color in zip(self.string_chars, self.string_colors)
        ]

    # change the gradient
    def update_gradient(self, gradient):
        self.__gradient = [c for c in gradient for _ in range(self.__gradient_length)]
        return self

    def generate(self, frame_number: int):
        if self.done_generating:
            return
        # is it time to change the noise chars?
        if frame_number % self.__char_halt == 0:
            self.__char_frame_number += 1
            for i, c in enumerate(self.string_chars):
                # frame == 0 basically
                if not c:
                    self.string_chars[i] = random.choice(
                        string.ascii_letters + "1234567890!@#$%^&*()_+-=<>,.:\";'{}[]?/"
                    )
                # randomly decide to update this char to a new one
                elif random.random() < 0.6:
                    self.string_chars[i] = random.choice(
                        string.ascii_letters + "1234567890!@#$%^&*()_+-=<>,.:\";'{}[]?/"
                    )
        # is it time to change the gradient position?
        if frame_number % self.__color_halt == 0:
            self.__color_frame_number += 1
            self.string_colors = [
                self.__gradient[(i - self.__color_frame_number) % len(self.__gradient)]
                for i in range(self.length)
            ]

        # update the color characters exposed to the main program
        self.colored_chars = [
            bruhcolored(c, color=color)
            for c, color in zip(self.string_chars, self.string_colors)
        ]

    def mark_done(self):
        self.done_generating = True


class Loading:
    def __init__(self, animate_part: GradientNoise):
        self.animate_part = animate_part

    def update(self, frame: int):
        self.animate_part.generate(frame)

    def mark_done(self):
        self.animate_part.mark_done()


class StringStreamer:
    def __init__(self, x: int, y: int, text: str, start_frame: int, halt: int = 1):
        self.x = x
        self.y = y
        self.text = text
        self.__start_frame = start_frame
        self.__halt = halt
        self.__chars = list(self.text)
        self.elapsed = []
        self.complete = False

    def generate(self, frame: int):
        if self.complete or frame < self.__start_frame:
            return
        if frame % self.__halt == 0:
            self.elapsed.append(self.__chars[len(self.elapsed)])
            if len(self.elapsed) == len(self.__chars):
                self.complete = True


class OllamaApiCaller:
    def __init__(
        self,
        model: str,
        use_message_history: bool = False,
        message_history_cap: int = 5,
    ):
        self.model = model
        self.url = "http://127.0.0.1:11434/api/chat"
        self.busy = False
        self.response = None
        self.state = "ready"
        self.use_message_history = use_message_history
        self.message_history = []
        self.message_history_cap = message_history_cap

    def chat(
        self, message: str, user: str | None, previous_messages: list[str] | None = None
    ) -> str:
        self.busy = True
        self.state = "running"
        payload = {
            "model": self.model,
            "messages": (
                [{"role": "user", "content": message}]
                if not self.message_history
                else self.message_history
                + [{"role": "user", "content": message}]
            ),
            "stream": False,
        }

        response = requests.post(url=self.url, data=json.dumps(payload))

        self.response = response.json()["message"]["content"]
        self.busy = False
        self.state = "finished"

        if self.use_message_history:
            self.message_history += [
                {"role": "user", "content": message},
                {"role": "assistant", "content": self.response},
            ]
            if len(self.message_history) > self.message_history_cap:
                self.message_history_cap = self.message_history_cap[1:]


class OpenAiCaller:
    def __init__(
        self,
        client: openai.OpenAI | openai.AzureOpenAI,
        model: str,
        use_message_history: bool = False,
        message_history_cap: int = 5,
    ):
        self.client = client
        self.model = model
        self.busy = False
        self.response = None
        self.state = "ready"
        self.use_message_history = use_message_history
        self.message_history = []
        self.message_history_cap = message_history_cap

    def chat(self, message: str, user: str | None) -> str:
        self.busy = True
        self.state = "running"
        response = self.client.chat.completions.create(
            model=self.model,
            messages=(
                [{"role": "user", "content": message}]
                if not self.message_history
                else self.message_history
                + [{"role": "user", "content": message}]
            ),
            max_tokens=500,
            temperature=0.5,
        )

        self.response = response.choices[0].message.content
        self.busy = False
        self.state = "finished"

        if self.use_message_history:
            self.message_history += [
                {"role": "user", "content": message},
                {"role": "system", "content": self.response},
            ]
            if len(self.message_history) > self.message_history_cap:
                self.message_history_cap = self.message_history_cap[
                    len(self.message_history) - self.message_history_cap :
                ]


class ChatbotEffect(BaseEffect):
    def __init__(self, screen: Screen, buffer, back_buffer, background: str = " "):
        super(ChatbotEffect, self).__init__(buffer, background)
        self.back_buffer = back_buffer
        self.screen = screen
        self.turn = 0
        self.interface = None
        self.model = None
        self.chatbot = None
        self.show_stats = False
        self.last_chatbot_response = ""
        self.last_chatbot_response_words = []

        # manages how the buffer is updated . . .
        self.all_keys = {idx: [] for idx in range(screen.height)}
        self.last_key = None
        self.user_message = ""

        self.second_effect = None
        self.chat_thread = None
        self.chatbot_thinker = None
        self.total_processed_chatbot_words = 0
        self.current_chatbot_response_words_idx = 0
        self.chatbot_print_halt = 25
        self.global_current_y_idx = 0

        self.avatar_size = 10

        self.user_y_turn_start_idx = 0
        self.chat_y_turn_start_idx = 0
        self.current_top_y = 0
        self.current_bottom_y = self.screen.height - 1

        self.user_cursor_x_idx = self.avatar_size
        self.user_cursor_y_idx = 0

        # colors
        self.chatbot_text_color = 243
        self.chatbot_background_color = None
        self.user_text_color = 255
        self.user_background_color = None
        self.user_avatar_color = None
        self.user_avatar_text_color = 255
        self.chatbot_avatar_color = None
        self.chatbot_avatar_text_color = 255

        self.gradient_noise_char_halt = 1
        self.gradient_noise_color_halt = 1
        self.blink_halt = 20
        self.blink_color_one = 255
        self.blink_color_two = 232
        self.cursor_char_color = 255

        self.divider = False
        self.divider_character = "-"

        self.gradient_text_color = [21, 57, 93, 129, 165, 201, 165, 129, 93, 57]
        self.gradient_idx = 1
        self.gradient_mul = 1

        for ydx in range(screen.height):
            for _ in range(self.avatar_size):
                self.all_keys[ydx] = [
                    Key(" ", [ord(" ")], ord(" "), x=_, y=ydx)
                    for _ in range(self.avatar_size)
                ]

        self.message_history = []

        self.view_y_top = 0
        self.view_y_bottom = self.screen.height - 1
        self.avatar_placed = False

    def __expand_list(self, original_list: list[int|str], n: int, mul: int = 1):
        l = []
        for val in original_list:
            for _ in range(mul):
                l.append(val)
        v = math.ceil(n / len(l))
        new_list = l * v
        return new_list[:n]

    def set_chatbot_properties(
        self,
        interface: str | None,
        model: str,
        user: str | None = None,
        client: openai.OpenAI | openai.AzureOpenAI | None = None,
        use_message_history: bool = False,
        message_history_cap: int = 5,
    ):
        if interface:
            self.interface = interface
        if user:
            self.user = user
        self.model = model
        if interface == "ollama":
            self.chatbot = OllamaApiCaller(
                model=self.model,
                use_message_history=use_message_history,
                message_history_cap=message_history_cap,
            )
        elif interface == "openai":
            if not client:
                raise Exception(
                    "An OpenAI client object must be provided for interface of type 'openai'."
                )
            self.chatbot = OpenAiCaller(
                client=client,
                model=model,
                use_message_history=use_message_history,
                message_history_cap=message_history_cap,
            )

    def set_second_effect(self, effect: str):
        self.second_effect = effect

    def set_chatbot_print_halt(self, halt: int):
        self.chatbot_print_halt = halt

    def set_gradient_noise_halts(
        self, char_halt: int | None = None, color_halt: int | None = None
    ):
        if char_halt:
            self.gradient_noise_char_halt = char_halt
        if color_halt:
            self.gradient_noise_color_halt = color_halt

    def set_chatbot_user_colors(
        self,
        chatbot_text_color: int | str | None = None,
        chatbot_background_color: int | str | None = None,
        chatbot_avatar_color: int | str | None = None,
        chatbot_avatar_text_color: int | str | None = None,
        user_text_color: int | str | None = None,
        user_background_color: int | str | None = None,
        user_avatar_color: int | str | None = None,
        user_avatar_text_color: int | str | None = None,
    ):
        if chatbot_text_color:
            self.chatbot_text_color = chatbot_text_color
        if chatbot_background_color:
            self.chatbot_background_color = chatbot_background_color
        if chatbot_avatar_color:
            self.chatbot_avatar_color = chatbot_avatar_color
        if chatbot_avatar_text_color:
            self.chatbot_avatar_text_color = chatbot_avatar_text_color
        if user_text_color:
            self.user_text_color = user_text_color
        if user_background_color:
            self.user_background_color = user_background_color
        if user_avatar_color:
            self.user_avatar_color = user_avatar_color
        if user_avatar_text_color:
            self.user_avatar_text_color = user_avatar_text_color

    def set_avatar_properties(self, size: int):
        self.avatar_size = size
        self.user_cursor_x_idx = self.avatar_size
        for ydx in range(self.screen.height):
            self.all_keys[ydx] = [
                Key(" ", [ord(" ")], ord(" "), x=_, y=ydx)
                for _ in range(self.avatar_size)
            ]

    def set_chatbot_stats(self, show: bool = False):
        self.show_stats = show

    def set_chatbot_blink_halt(self, halt: int):
        self.blink_halt = halt

    def set_divider_flag(self, divider: bool, divider_character: str = "-"):
        self.divider = divider
        self.divider_character = divider_character

    def set_chatbot_cursor_colors(self, color_one: int | str, color_two: int | str):
        self.blink_color_one = color_one
        self.blink_color_two = color_two

    def set_chatbot_text_gradient(self, gradient: list[int|str], mul: int):
        self.gradient_text_color = gradient
        self.gradient_mul = mul

    def __handle_keyboard_result(self, result):
        if result:
            if result == -300:  # backspace
                if self.user_cursor_x_idx == self.avatar_size:
                    return
                all_key_history = self.all_keys[self.user_cursor_y_idx]
                if all_key_history == []:
                    if self.user_cursor_y_idx - 1 >= self.global_current_y_idx:
                        self.user_cursor_y_idx -= 1
                        if self.user_cursor_y_idx in self.all_keys:
                            keys = self.all_keys[self.user_cursor_y_idx]
                            if keys:
                                self.user_cursor_x_idx = self.all_keys[
                                    self.user_cursor_y_idx
                                ][-1].x
                            else:
                                self.user_cursor_x_idx = 0
                        else:
                            self.user_cursor_x_idx = 0
                else:
                    last_key = all_key_history[-1]
                    self.user_cursor_x_idx -= len(last_key.representation)
                    self.all_keys[self.user_cursor_y_idx] = all_key_history[:-1]
                if self.user_cursor_y_idx < 0:
                    self.user_cursor_y_idx = 0
                if len(self.user_message) > 0:
                    self.user_message = self.user_message[:-1]
                return False
            elif result == 10:  # enter
                self.user_cursor_y_idx += 1
                if self.user_cursor_y_idx >= max(self.all_keys):
                    self.scroll_keys(shift=1)
                self.user_cursor_x_idx = self.avatar_size
                return self.last_key != "\\"
            elif result == 11:  # tab
                self.user_cursor_x_idx += 4
                self.all_keys[self.user_cursor_y_idx].append(
                    Key(
                        character=" ",
                        representation=[32, 32, 32, 32],
                        value=result,
                        x=self.user_cursor_x_idx,
                        y=self.user_cursor_y_idx,
                    )
                )
                return False
            elif result == -204: # arrow up
                if self.current_top_y != 0:
                    self.current_top_y -= 1
                    self.current_bottom_y -= 1
            elif result == -206: # arrow down
                if self.current_bottom_y < (len(self.all_keys) - 1):
                    self.current_top_y += 1
                    self.current_bottom_y += 1
            elif result > 0:
                self.user_cursor_x_idx += 1
                self.all_keys[self.user_cursor_y_idx].append(
                    Key(
                        character=bruhcolored(
                            text=chr(result),
                            color=self.user_text_color,
                            on_color=self.user_background_color,
                        ).colored,
                        representation=[result],
                        value=result,
                        x=self.user_cursor_x_idx,
                        y=self.user_cursor_y_idx,
                    )
                )
                self.last_key = chr(result)
                self.user_message += chr(result)
                return False

    def __scroll_up(self):
        first_key = min(self.all_keys.keys())
        last_key = max(self.all_keys.keys())
        if len(self.all_keys) == 0 or first_key == 0:
            return
        for key in range(first_key, last_key):
            self.all_keys[key] = self.all_keys[key + 1]
        self.all_keys[last_key] = [
                    Key(" ", [ord(" ")], ord(" "), x=_, y=last_key)
                    for _ in range(self.avatar_size)
                ]

        if self.current_top_y > 0:
            self.current_top_y = self.current_top_y - 1
            self.current_bottom_y = self.current_bottom_y - 1

    def __scroll_down(self):
        last_key = max(self.all_keys.keys())
        self.all_keys[last_key + 1] = [
                    Key(" ", [ord(" ")], ord(" "), x=_, y=last_key + 1)
                    for _ in range(self.avatar_size)
                ]
        # if len(self.all_keys) >= 100:
        #     first_key = min(self.all_keys.keys())
        #     del self.all_keys[first_key]
        #     self.all_keys = {k-1:v for k,v in self.all_keys.items()}
        # else:
        self.current_top_y = self.current_top_y + 1
        self.current_bottom_y = self.current_bottom_y + 1
        self.all_keys = {i: self.all_keys[key] for i, key in enumerate(sorted(self.all_keys.keys()))}

    def scroll_keys(self, shift: int = 1):
        self.__scroll_down() if shift == 1 else self.__scroll_up()

    def place_all_keys(self):
        for idx, vals in self.all_keys.items():
            if idx > self.current_bottom_y or idx < self.current_top_y:
                continue
            displacement = 0
            for jdx, key in enumerate(vals):
                if len(key.representation) > 1:
                    for val in key.representation:
                        self.buffer.put_char(jdx + displacement, idx - self.current_top_y, chr(val))
                        displacement += 1
                    displacement -= 1
                else:
                    self.buffer.put_char(jdx + displacement, idx - self.current_top_y, key.character)

    def render_frame(self, frame_number):
        self.buffer.clear_buffer()
        if self.turn == 0:
            if not self.avatar_placed:
                for i, c in enumerate(self.user[: self.avatar_size - 1]):
                    self.all_keys[self.user_y_turn_start_idx][i] = Key(
                        bruhcolored(
                            c,
                            color=self.user_avatar_text_color,
                            on_color=self.user_avatar_color,
                        ).colored,
                        [ord(c)],
                        ord(c),
                        None,
                        None,
                    )
                for i in range(len(self.user), self.avatar_size - 1):
                    self.all_keys[self.user_y_turn_start_idx][i] = Key(
                        bruhcolored(" ", on_color=self.user_avatar_color).colored,
                        [ord(" ")],
                        ord(" "),
                        None,
                        None,
                    )
                self.avatar_placed = True
            self.screen.wait_for_input(timeout=0)
            result = self.screen.get_event()
            flip_turn = self.__handle_keyboard_result(result=result)
            if flip_turn:
                self.avatar_placed = False
                self.turn = 1
                if self.divider:
                    for _ in range(self.avatar_size):
                        self.all_keys[self.user_cursor_y_idx].append(
                            Key(
                                self.divider_character,
                                [ord(self.divider_character)],
                                ord(self.divider_character),
                                None,
                                None,
                            )
                        )
                    for _ in range(self.avatar_size, self.buffer.width()):
                        self.all_keys[self.user_cursor_y_idx].append(
                            Key(
                                self.divider_character,
                                [ord(self.divider_character)],
                                ord(self.divider_character),
                                None,
                                None,
                            )
                        )
                    self.user_cursor_y_idx += 1
                    if self.user_cursor_y_idx >= max(self.all_keys):
                        self.scroll_keys(shift=1)
                self.chat_y_turn_start_idx = self.user_cursor_y_idx
        else:
            # call llm with self.user_message
            if self.chatbot.state == "ready":
                message = self.user_message[:]
                self.thread = Thread(
                    target=self.chatbot.chat,
                    args=(
                        message,
                        self.user,
                    ),
                )
                self.thread.start()
                self.chatbot_thinker = GradientNoise(
                    x=0,
                    y=self.user_cursor_y_idx,
                    length=self.avatar_size - 1,
                    char_halt=self.gradient_noise_char_halt,
                    color_halt=self.gradient_noise_color_halt,
                    gradient_length=2,
                )  # .update_gradient([21, 57, 93, 129, 165, 201, 165, 129, 93, 57])
            elif (
                self.thread
                and not self.thread.is_alive()
                and self.chatbot.state == "finished"
            ):
                self.thread.join()
                self.thread = None
                if len(self.chatbot.response) > self.buffer.width():
                    self.last_chatbot_response = self.chatbot.response
                else:
                    self.last_chatbot_response = self.chatbot.response
                self.last_chatbot_response_words = self.last_chatbot_response.split(" ")
                self.chatbot.response = ""
                self.user_message = ""
                self.chatbot.state = "printing"
                if self.chatbot_text_color == "gradient":
                    self.gradient_idx = 1
                    self.gradient_text_color_for_message = self.__expand_list(self.gradient_text_color, len(self.last_chatbot_response), 3)
            elif self.chatbot.state == "printing":
                # print out the characters!
                if not self.avatar_placed:
                    for i, c in enumerate(self.model[: self.avatar_size - 1]):
                        self.all_keys[self.chat_y_turn_start_idx][i] = Key(
                            bruhcolored(
                                c,
                                color=self.chatbot_avatar_text_color,
                                on_color=self.chatbot_avatar_color,
                            ).colored,
                            [ord(c)],
                            ord(c),
                            None,
                            None,
                        )
                    for i in range(len(self.model), self.avatar_size - 1):
                        self.all_keys[self.chat_y_turn_start_idx][i] = Key(
                            bruhcolored(" ", on_color=self.chatbot_avatar_color).colored,
                            [ord(" ")],
                            ord(" "),
                            None,
                            None,
                        )
                    self.avatar_placed = True
                if frame_number & self.chatbot_print_halt == 0:
                    if self.total_processed_chatbot_words == len(self.last_chatbot_response_words):
                        self.chatbot.state = "done"
                        self.current_chatbot_response_words_idx = 0
                        self.total_processed_chatbot_words = 0
                    else:
                        next_word = (
                            self.last_chatbot_response_words[
                                self.current_chatbot_response_words_idx
                            ]
                            + " "
                        )
                        for idx, character in enumerate(next_word):
                            if self.user_cursor_y_idx >= max(self.all_keys):
                                self.scroll_keys(shift=1)
                            if character == "\n":
                                self.user_cursor_y_idx += 1
                                if self.user_cursor_y_idx >= max(self.all_keys):
                                    self.scroll_keys(shift=1)
                                self.all_keys[self.user_cursor_y_idx][
                                    self.avatar_size - 2
                                ] = Key(
                                    character=">",
                                    representation=[ord(">")],
                                    value=ord(">"),
                                    x=None,
                                    y=None,
                                )
                                continue
                            self.all_keys[self.user_cursor_y_idx].append(
                                Key(
                                    character=bruhcolored(
                                        text=character,
                                        color=self.chatbot_text_color if self.chatbot_text_color != "gradient" else self.gradient_text_color_for_message[self.gradient_idx - 1],
                                        on_color=self.chatbot_background_color,
                                    ).colored,
                                    representation=[ord(character)],
                                    value=ord(character),
                                    x=self.user_cursor_x_idx,
                                    y=self.user_cursor_y_idx,
                                )
                            )
                            if character != " ":
                                self.gradient_idx += 1
                            if (
                                len(self.all_keys[self.user_cursor_y_idx])
                                == self.buffer.width()
                            ):
                                self.user_cursor_y_idx += 1
                                if self.user_cursor_y_idx >= max(self.all_keys):
                                    self.scroll_keys(shift=1)
                                self.all_keys[self.user_cursor_y_idx][
                                    self.avatar_size - 2
                                ] = Key(
                                    character=">",
                                    representation=[ord(">")],
                                    value=ord(">"),
                                    x=None,
                                    y=None,
                                )
                        self.current_chatbot_response_words_idx += 1
                        self.total_processed_chatbot_words += 1
            elif self.chatbot.state == "done":
                self.turn = 0
                self.avatar_placed = False
                self.chatbot_thinker = None
                self.chatbot.state = "ready"
                self.global_current_y_idx += 1
                if self.divider:
                    self.user_cursor_y_idx += 1
                    if self.user_cursor_y_idx >= max(self.all_keys):
                        self.scroll_keys(shift=1)
                    for _ in range(self.avatar_size):
                        self.all_keys[self.user_cursor_y_idx].append(
                            Key(
                                self.divider_character,
                                [ord(self.divider_character)],
                                ord(self.divider_character),
                                None,
                                None,
                            )
                        )
                    for _ in range(self.avatar_size, self.buffer.width()):
                        self.all_keys[self.user_cursor_y_idx].append(
                            Key(
                                self.divider_character,
                                [ord(self.divider_character)],
                                ord(self.divider_character),
                                None,
                                None,
                            )
                        )
                self.user_cursor_y_idx += 1
                if self.user_cursor_y_idx >= max(self.all_keys):
                    self.scroll_keys(shift=1)
                self.user_y_turn_start_idx = self.user_cursor_y_idx
            else:
                self.chatbot_thinker.generate(frame_number=frame_number)

        if self.second_effect:
            self.second_effect.render_frame(frame_number=frame_number)
            self.buffer.sync_with(self.second_effect.buffer)

        self.place_all_keys()

        if self.chatbot_thinker and self.chatbot.state == "running":
            for i, c in enumerate(self.chatbot_thinker.colored_chars):
                self.buffer.put_char(
                    self.chatbot_thinker.x + i, self.chatbot_thinker.y, c.colored
                )

        if self.show_stats:
            self.buffer.put_at(
                0,
                self.buffer.height() - 5,
                f"CURSOR X: {self.user_cursor_x_idx}",
                transparent=True,
            )
            self.buffer.put_at(
                0,
                self.buffer.height() - 4,
                f"CURSOR Y: {self.user_cursor_y_idx}",
                transparent=True,
            )
            self.buffer.put_at(
                0,
                self.buffer.height() - 3,
                f"CHATBOT STATE: {self.chatbot.state}",
                transparent=True,
            )
            self.buffer.put_at(
                0,
                self.buffer.height() - 2,
                f"CHATBOT RESPONSE WORD COUNT: {0 if not self.last_chatbot_response_words else len(self.last_chatbot_response_words)}",
                transparent=True,
            )
            self.buffer.put_at(
                0,
                self.buffer.height() - 1,
                f"PROCESSED RESPONSE WORDS: {self.total_processed_chatbot_words}",
                transparent=True,
            )

        if self.turn == 0:
            if frame_number % self.blink_halt == 0:
                self.cursor_char_color = (
                    self.blink_color_one
                    if self.cursor_char_color == self.blink_color_two
                    else self.blink_color_two
                )
            self.buffer.put_char(
                self.user_cursor_x_idx,
                self.user_cursor_y_idx,
                bruhcolored(" ", on_color=self.cursor_char_color).colored,
                transparent=False,
            )



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .base_effect import BaseEffect


class Line:
    def __init__(self, start_point, end_point):
        if start_point and end_point:
            self.start_point = (start_point[0] * 2, start_point[1] * 2)
            self.end_point = (end_point[0] * 2, end_point[1] * 2)
        else:
            self.start_point = None
            self.end_point = None

    def update_points(self, start_point, end_point):
        self.start_point = (start_point[0], start_point[1])
        self.end_point = (end_point[0], end_point[1])

    def get_points(self):
        return self.start_point, self.end_point


class DrawLinesEffect(BaseEffect):
    def __init__(self, buffer, background, char=None, thin=False):
        super(DrawLinesEffect, self).__init__(buffer, background)
        self.lines = []
        self.char = char
        self.thin = thin

    def add_line(self, start_point, end_point):
        self.lines.append(Line(start_point, end_point))

    def render_frame(self, frame_number):
        if frame_number == 0 and len(self.lines) > 0:
            for y in range(self.buffer.height()):
                self.buffer.put_at(0, y, self.background * self.buffer.width())
            for line in self.lines:
                if (
                    (line.start_point[0] < 0 and line.end_point[0]) < 0
                    or (
                        line.start_point[0] >= self.buffer.width() * 2
                        and line.end_point[0] > self.buffer.width() * 2
                    )
                    or (line.start_point[1] < 0 and line.end_point[1] < 0)
                    or (
                        line.start_point[1] >= self.buffer.height() * 2
                        and line.end_point[1] >= self.buffer.height() * 2
                    )
                ):
                    return

                line_chars = " ''^.|/7.\\|Ywbd#"
                dx = abs(line.end_point[0] - line.start_point[0])
                dy = abs(line.end_point[1] - line.start_point[1])

                cx = -1 if line.start_point[0] > line.end_point[0] else 1
                cy = -1 if line.start_point[1] > line.end_point[1] else 1

                def get_start(x, y):
                    c = self.buffer.get_char(x, y)
                    if c is not None:
                        return line_chars.find(c)
                    return 0

                def x_draw(ix, iy):
                    err = dx
                    px = ix - 2
                    py = iy - 2
                    next_char = 0
                    while ix != line.end_point[0]:
                        if ix < px or ix - px >= 2 or iy < py or iy - py >= 2:
                            px = ix & ~1
                            py = iy & ~1
                            next_char = get_start(px // 2, py // 2)
                        next_char |= 2 ** abs(ix % 2) * 4 ** (iy % 2)
                        err -= 2 * dy
                        if err < 0:
                            iy += cy
                            err += 2 * dx
                        ix += cx

                        if self.char is None:
                            self.buffer.put_char(
                                px // 2, py // 2, line_chars[next_char]
                            )
                        else:
                            self.buffer.put_char(px // 2, py // 2, self.char)

                def y_draw(ix, iy):
                    err = dy
                    px = ix - 2
                    py = iy - 2
                    next_char = 0

                    while iy != line.end_point[1]:
                        if ix < px or ix - px >= 2 or iy < py or iy - py >= 2:
                            px = ix & ~1
                            py = iy & ~1
                            next_char = get_start(px // 2, py // 2)
                        next_char |= 2 ** abs(ix % 2) * 4 ** (iy % 2)
                        err -= 2 * dx
                        if err < 0:
                            ix += cx
                            err += 2 * dy
                        iy += cy

                        if self.char is None:
                            self.buffer.put_char(
                                px // 2, py // 2, line_chars[next_char]
                            )
                        else:
                            self.buffer.put_char(px // 2, py // 2, self.char)

                if dy == 0 and self.thin and self.char is None:
                    pass
                elif dx > dy:
                    x_draw(line.start_point[0], line.start_point[1] + 1)
                else:
                    y_draw(line.start_point[0] + 1, line.start_point[1])



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List
from .base_renderer import BaseRenderer
from bruhcolor import bruhcolored
from ..bruhutil.bruhtypes import EffectType


class BackgroundColorRenderer(BaseRenderer):
    def __init__(
        self,
        screen,
        img: List[str],
        frames: int = 100,
        frame_time: float = 0.1,
        effect_type: EffectType = "static",
        background: str = " ",
        transparent: bool = False,
        collision: bool = False,
        on_color_code: int = 27,
    ):
        super(BackgroundColorRenderer, self).__init__(
            screen, frames, frame_time, effect_type, background, transparent, collision
        )

        self.img = img
        self.img_height = len(self.img)
        self.img_width = len(self.img[0])
        self.img_y_start = (self.height - len(self.img)) // 2
        self.img_x_start = (self.width - len(self.img[0])) // 2
        self.current_img_x = self.img_x_start
        self.current_img_y = self.img_y_start

        if not on_color_code:
            raise Exception("a color code must be provided to BackgroundColorRenderer")
        if (
            not isinstance(on_color_code, int)
            or on_color_code < 0
            or on_color_code > 255
        ):
            raise Exception("the color code must be an int value 0-255")
        self.on_color_code = on_color_code

    def render_img_frame(self, frame_number):
        for y in range(self.height):
            for x in range(self.width):
                if (
                    (y >= self.img_y_start)
                    and (y < (self.img_y_start + self.img_height))
                    and (x >= self.img_x_start)
                    and (x < (self.img_x_start + self.img_width))
                ):
                    self.image_buffer.put_char(
                        x,
                        y,
                        bruhcolored(
                            self.img[y - self.img_y_start][x - self.img_x_start],
                            on_color=self.on_color_code,
                        ).colored,
                    )



from .base_renderer import BaseRenderer
from .effect_renderer import EffectRenderer
from .center_renderer import CenterRenderer
from .pan_renderer import PanRenderer
from .focus_renderer import FocusRenderer
from .background_color_renderer import BackgroundColorRenderer


__all__ = [
    "BaseRenderer",
    "EffectRenderer",
    "CenterRenderer",
    "PanRenderer",
    "FocusRenderer",
    "BackgroundColorRenderer",
]



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import random
import string

from bruhcolor import bruhcolored
from .base_effect import BaseEffect


class MatrixEffect(BaseEffect):
    """
    Effect to mimic the cliche coding backgroud with falling random characters
    """

    def __init__(
        self,
        buffer,
        background,
        chracter_halt_range=(1, 2),
        color_halt_range=(1, 2),
        character_randomness_one=0.70,
        character_randomness_two=0.60,
        color_randomness=0.50,
        gradient_length=1,
    ):
        super(MatrixEffect, self).__init__(buffer, background)
        self.__character_choices = (
            string.ascii_letters + "1234567890!@#$%^&*()_+-=<>,.:\";'{}[]?/"
        )
        self.__character_halt_range = chracter_halt_range
        self.__color_halt_range = color_halt_range
        self.__character_halts = [
            random.randint(
                self.__character_halt_range[0], self.__character_halt_range[1]
            )
            for _ in range(self.buffer.height())
        ]
        self.__color_halts = [
            random.randint(self.__color_halt_range[0], self.__color_halt_range[1])
            for _ in range(self.buffer.height())
        ]
        self.__character_randomness_one = character_randomness_one
        self.__character_randomness_two = character_randomness_two
        self.__color_randomness = color_randomness
        self.__gradient_length = gradient_length
        self.__base_gradient = [
            232,
            233,
            234,
            235,
            236,
            237,
            238,
            239,
            240,
            241,
            242,
            243,
            244,
            245,
            246,
            247,
            248,
            249,
            250,
            251,
            252,
            253,
            254,
            255,
        ]
        self.__gradient = [
            color
            for color in self.__base_gradient
            for _ in range(self.__gradient_length)
        ]
        self.__character_frame_numbers = [0 for _ in range(self.buffer.height())]
        self.__color_frame_numbers = [0 for _ in range(self.buffer.height())]
        self.__buffer_characters = [
            [" " for x in range(self.buffer.width())]
            for y in range(self.buffer.height())
        ]

    def set_matrix_properties(
        self,
        chacter_halt_range=(1, 2),
        color_halt_range=(1, 2),
        character_randomness_one=0.70,
        character_randomness_two=0.60,
        color_randomness=0.50,
        gradient_length=1,
    ):
        self.__character_randomness_one = character_randomness_one
        self.__character_randomness_two = character_randomness_two
        self.__color_randomness = color_randomness
        self.__character_halt_range = chacter_halt_range
        self.__color_halt_range = color_halt_range
        self.__gradient_length = gradient_length
        self.__gradient = [
            color
            for color in self.__base_gradient
            for _ in range(self.__gradient_length)
        ]
        self.__character_halts = [
            random.randint(
                self.__character_halt_range[0], self.__character_halt_range[1]
            )
            for _ in range(self.buffer.height())
        ]
        self.__color_halts = [
            random.randint(self.__color_halt_range[0], self.__color_halt_range[1])
            for _ in range(self.buffer.height())
        ]

    def set_matrix_gradient(self, gradient):
        self.__base_gradient = gradient
        self.__gradient = [
            color
            for color in self.__base_gradient
            for _ in range(self.__gradient_length)
        ]

    def get_gradient(self):
        return self.__base_gradient

    def __initialize_buffer(self):
        for y in range(self.buffer.height()):
            for x in range(self.buffer.width()):
                self.__buffer_characters[y][x] = random.choice(self.__character_choices)
            for x in range(self.buffer.width()):
                self.buffer.put_char(
                    x,
                    y,
                    bruhcolored(
                        self.__buffer_characters[y][x],
                        self.__gradient[x % len(self.__gradient)],
                    ).colored,
                )

    def render_frame(self, frame_number):
        """
        Renders the next frame for the Matrix effect into the effect buffer
        """
        if frame_number == 0:
            self.__initialize_buffer()
        else:
            for y in range(self.buffer.height()):
                if (
                    frame_number % self.__character_halts[y] == 0
                    and random.random() < self.__character_randomness_one
                ):
                    self.__character_frame_numbers[y] += 1
                    for x in range(self.buffer.width()):
                        if random.random() < self.__character_randomness_two:
                            self.__buffer_characters[y][x] = random.choice(
                                self.__character_choices
                            )

                if (
                    frame_number % self.__color_halts[y] == 0
                    and random.random() < self.__color_randomness
                ):
                    self.__color_frame_numbers[y] += 1
                    for x in range(self.buffer.width()):
                        self.buffer.put_char(
                            x,
                            y,
                            bruhcolored(
                                self.__buffer_characters[y][x],
                                color=self.__gradient[
                                    (x - self.__color_frame_numbers[y])
                                    % len(self.__gradient)
                                ],
                            ),
                        )



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
from .base_renderer import BaseRenderer
from ..bruhutil import Screen, INF, sleep
from ..bruhutil.bruherrors import ScreenResizedError
from ..bruhutil.bruhtypes import EffectType


class EffectRenderer(BaseRenderer):
    """
    Class for rendering the Effect and only the Effect
    """

    def __init__(

        self,
        screen: Screen,
        frames: int = 100,
        frame_time: float = 0.1,
        effect_type: EffectType = "static",
        background: str = " ",
        transparent: bool = False,
    ):
        super(EffectRenderer, self).__init__(
            screen, frames, frame_time, effect_type, background, transparent
        )

        self.background = self.effect.background

    def render_effect_frame(self, frame_number: int):
        """
        We only need to render the effect, so we just call the effects render
        frame method to update the effect buffer
        """
        self.effect.render_frame(frame_number)

    def run(self, end_message: bool = True):
        """
        Generate the next effect frame and sync it with the back / front buffer
        """
        try:
            if self.frames == INF:
                frame = 0
                while True:
                    if self.screen.has_resized(): raise ScreenResizedError("An error was encounter. The Screen was resized.")
                    sleep(self.frame_time)
                    self.render_effect_frame(frame)
                    self.back_buffer.sync_with(self.effect.buffer)
                    self.push_front_to_screen()
                    self.front_buffer.sync_with(self.back_buffer)
                    frame += 1
            else:
                for frame in range(self.frames):
                    if self.screen.has_resized(): raise ScreenResizedError("An error was encounter. The Screen was resized.")
                    sleep(self.frame_time)
                    self.render_effect_frame(frame)
                    self.back_buffer.sync_with(self.effect.buffer)
                    self.push_front_to_screen()
                    self.front_buffer.sync_with(self.back_buffer)
            if end_message:
                self.render_exit()
                self.push_front_to_screen()
            if sys.platform == 'win32': input()
        except KeyboardInterrupt:
            if end_message:
                self.render_exit()
                self.push_front_to_screen()
            if sys.platform == 'win32': input()

        if end_message:
            self.render_exit()
            self.push_front_to_screen()



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, Tuple, Optional
from .base_renderer import BaseRenderer
from ..bruhutil import Screen
from ..bruhutil.bruhtypes import EffectType, PanRendererDirection, valid_pan_renderer_directions
from ..bruhutil.bruherrors import InvalidPanRendererDirectionError


class PanRenderer(BaseRenderer):
    """
    A renderer to pan an image across the screen.
    Updates the image_buffer only.
    """

    def __init__(
        self,
        screen: Screen,
        img: List[str],
        frames: int = 100,
        frame_time: float = 0.1,
        effect_type: EffectType = "static",
        background: str = " ",
        transparent: bool = False,
        collision: bool = False,
        direction: PanRendererDirection = "horizontal",
        shift_rate: int = 1,
        loop: bool = False,
    ) -> None:
        super().__init__(screen, frames, frame_time, effect_type, background, transparent, collision)
        self.direction = self.validate_direction(direction)
        self.img = img
        self.shift_rate = max(1, int(shift_rate))
        self.loop = loop
        if self.img:
            self._set_img_attributes()

    def validate_direction(self, direction: PanRendererDirection) -> PanRendererDirection:
        if direction not in valid_pan_renderer_directions:
            raise InvalidPanRendererDirectionError(
                f"Invalid direction for PanRenderer. Please choose from {valid_pan_renderer_directions}"
            )
        return direction

    def _set_img_attributes(self) -> None:
        """Sets the attributes for the image given it exists."""
        self.render_image = True
        self.img_height = len(self.img)
        self.img_width = len(self.img[0])
        self.img_back = -self.img_width - 1
        self.img_front = -1
        self.img_top = (self.height - self.img_height) // 2
        self.img_bottom = ((self.height - self.img_height) // 2) + self.img_height
        self.current_img_x = self.img_back
        self.current_img_y = self.img_top

    @property
    def img_size(self) -> Tuple[int, int]:
        """Return the current image dimensions."""
        return len(self.img), len(self.img[0]) if self.img else (0, 0)

    def render_img_frame(self, frame_number: int) -> None:
        """Renders out the next frame of the pan animation."""
        if not self.loop and self.img_back > self.width + 1:
            return

        if self.direction == "horizontal":
            self.render_horizontal_frame(frame_number=frame_number)
        elif self.direction == "vertical":
            self.render_vertical_frame(frame_number=frame_number)

    def _set_padding(self, padding_vals: Tuple[int, int]) -> None:
        """Set the padding for the image [DEPRECATED FOR NOW]."""
        if not self.img or len(padding_vals) != 2:
            return

        left_right, top_bottom = padding_vals
        self.padding = (left_right, top_bottom)

        self.img = [" " * self.img_width for _ in range(top_bottom)]
        + [(" " * left_right) + line + (" " * left_right) for line in self.img]
        + [" " * self.img_width for _ in range(top_bottom)]

        self._set_img_attributes()

    def render_horizontal_frame(self, frame_number):
        """
        Renders the next image frame for a horizontal pan
        """
        if self.shift_rate > 0:
            if (
                0 <= frame_number <= self.img_width // self.shift_rate + 1
            ) or not self.loop:
                for y in range(self.height):
                    for x in range(self.width):
                        if (
                            x >= self.img_back
                            and x < self.img_front
                            and y >= self.img_top
                            and y < self.img_bottom
                        ):
                            if (
                                (y - self.img_top) >= 0
                                and (y - self.img_bottom) < self.img_height
                                and (x - self.img_back) >= 0
                                and (x - self.img_back) < self.img_width
                            ):
                                if self.transparent:
                                    if (
                                        self.img[y - self.img_top][x - self.img_back]
                                        == " "
                                    ):
                                        self.image_buffer.put_char(x, y, None)
                                    else:
                                        self.image_buffer.put_char(
                                            x,
                                            y,
                                            self.img[y - self.img_top][
                                                x - self.img_back
                                            ],
                                        )
                                else:
                                    self.image_buffer.put_char(
                                        x,
                                        y,
                                        self.img[y - self.img_top][x - self.img_back],
                                    )
                        else:
                            self.image_buffer.put_char(x, y, None)
                if self.loop:
                    if self.img_front >= self.width:
                        self.img_front = 0
                    else:
                        self.img_front += self.shift_rate
                    if self.img_back >= self.width:
                        self.img_back = 0
                    else:
                        self.img_back += self.shift_rate
                else:
                    self.img_back += self.shift_rate
                    self.img_front += self.shift_rate
            else:
                self.image_buffer.shift(-self.shift_rate)
        else:
            pass

    def render_vertical_frame(self, frame_number):
        pass



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .base_renderer import BaseRenderer
from ..bruhutil import Screen
from ..bruhutil.bruhtypes import EffectType


class CenterRenderer(BaseRenderer):
    """
    A renderer to load an image in the center of the screen.
    Updates the image_buffer only
    """

    def __init__(
        self,
        screen: Screen,
        img: list[str],
        frames: int = 100,
        frame_time: float = 0.1,
        effect_type: EffectType = "static",
        background: str = " ",
        transparent: bool = False,
        collision: bool = False
    ):
        super(CenterRenderer, self).__init__(
            screen, frames, frame_time, effect_type, background, transparent, collision
        )
        self.background = background
        self.transparent = transparent

        # Image attributes
        self.img = img
        self.img_height = len(self.img)
        self.img_width = len(self.img[0])
        self.img_y_start = (self.height - self.img_height) // 2
        self.img_x_start = (self.width - self.img_width) // 2
        self.current_img_x = self.img_x_start
        self.current_img_y = self.img_y_start
        self.none_fill_char = None

    def render_img_frame(self, frame_number):
        """
        Renders out the image to the center of the screen,
        if there is no image passed into the renderer then
        the background is rendered on it's own
        """

        # Image is only rendered once, on frame 0
        if frame_number == 0:
            if self.smart_transparent:
                for y in range(self.height):
                    for x in range(self.width):
                        if (
                            y >= self.img_y_start
                            and y < self.img_y_start + self.img_height
                            and x >= self.img_x_start
                            and x < self.img_x_start + self.img_width
                        ):
                            self.image_buffer.put_char(
                                x,
                                y,
                                self.img[y - self.img_y_start][x - self.img_x_start],
                            )
                # Now process spaces from left-to-right till a non-space character is hit.
                # Then do the same right-to-left. Place these spaces with None
                for y in range(self.height):
                    if y >= self.img_y_start and y < self.img_y_start + self.img_height:
                        for x in range(self.width):
                            if (
                                x >= self.img_x_start
                                and x < self.img_x_start + self.img_width
                            ):
                                if self.image_buffer.get_char(x, y) != " ":
                                    break
                                else:
                                    self.image_buffer.put_char(x, y, None)
                            else:
                                self.image_buffer.put_char(x, y, None)
                        for x in range(self.width - 1, -1, -1):
                            if (
                                x >= self.img_x_start
                                and x < self.img_x_start + self.img_width
                            ):
                                if self.image_buffer.get_char(x, y) != " ":
                                    break
                                else:
                                    self.image_buffer.put_char(x, y, None)
                            else:
                                self.image_buffer.put_char(x, y, None)
                    else:
                        for x in range(self.width):
                            self.image_buffer.put_char(x, y, None)
            else:
                for y in range(self.height):
                    if y >= self.img_y_start and y < self.img_y_start + self.img_height:
                        self.image_buffer.put_at_center(
                            y,
                            self.img[y - self.img_y_start],
                            transparent=self.transparent,
                        )


"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import math
import random

from bruhcolor import bruhcolored
from ..bruhutil import PLASMA_COLORS, GREY_SCALES
from .base_effect import BaseEffect


class PlasmaEffect(BaseEffect):
    """
    Function to generate a plasma like effect
    """

    def __init__(self, buffer, background):
        super(PlasmaEffect, self).__init__(buffer, background)
        self.info = False
        self.random = False
        self.scale = random.choice(GREY_SCALES)
        self.ayo = 0
        self.color = False
        self.vals = [
            random.randint(1, 100),
            random.randint(1, 100),
            random.randint(1, 100),
            random.randint(1, 100),
        ]

    def update_info_visibility(self, visible):
        """
        Function to enable or disable info about the effect
        """
        self.info = visible

    def update_grey_scale_size(self, size):
        """
        Function to change the size of the grey scale
        """
        if size in [8, 10, 16]:
            self.scale = random.choice(
                [scale for scale in GREY_SCALES if len(scale) == size]
            )
            if not self.random:
                self.colors = random.choice(PLASMA_COLORS[size])
            else:
                self.colors = [random.randint(0, 255) for _ in range(len(self.scale))]
        else:
            raise Exception(
                f"only 8, 10, and 16 are supported grey scale sizes, you provided {size}"
            )

    def update_color_properties(self, color, characters=True, random_color=False):
        """
        Function to update the color properties. random_color overrules other functions
        like update greyscale size and update color
        :param color: True / False to enable color
        :param characters: True / False to show the characters
        :param random_color: True / False to generate random colors
        """
        self.color = color
        self.random = random_color
        if not random_color:
            self.colors = PLASMA_COLORS[len(self.scale)][0]
        else:
            self.colors = [random.randint(0, 255) for _ in range(len(self.scale))]
        self.characters = characters

    def update_color(self, colors):
        """
        Function to update the colors used
        """
        if not self.random:
            if len(colors) == len(self.scale):
                self.colors = colors
            else:
                raise Exception(
                    f"update_color(..) must be provided a list of {len(self.scale)} colors, you provided {len(colors)}"
                )

    def update_background(self, background):
        """
        Update the background character(s)
        :param background: the new background
        """
        self.background = background
        self.background_length = len(self.background)

    def update_plasma_values(
        self,
        a=random.randint(1, 50),
        b=random.randint(1, 50),
        c=random.randint(1, 50),
        d=random.randint(1, 50),
    ):
        """
        Function to set the plasma values
        """
        self.vals = [a, b, c, d]

    def shuffle_plasma_values(self):
        """
        Function to generate a new-random set of plasma values
        """
        self.vals = [
            random.randint(1, 50),
            random.randint(1, 50),
            random.randint(1, 50),
            random.randint(1, 50),
        ]

    def render_frame(self, frame_number):
        """
        Function to render the next frame of the Plasma Effect
        """
        self.ayo += 1
        for y in range(self.buffer.height()):
            for x in range(self.buffer.width()):
                value = (
                    abs(
                        self.func(x + self.ayo / 3, y, 1 / 4, 1 / 3, self.vals[0])
                        + self.func(x, y, 1 / 8, 1 / 5, self.vals[1])
                        + self.func(x, y + self.ayo / 3, 1 / 2, 1 / 5, self.vals[2])
                        + self.func(x, y, 3 / 4, 4 / 5, self.vals[3])
                    )
                    / 4.0
                )
                if self.color:
                    if self.characters:
                        self.buffer.put_char(
                            x,
                            y,
                            bruhcolored(
                                self.scale[int((len(self.scale) - 1) * value)],
                                color=self.colors[int((len(self.scale) - 1) * value)],
                            ).colored,
                        )
                    else:
                        self.buffer.put_char(
                            x,
                            y,
                            bruhcolored(
                                " ",
                                on_color=self.colors[
                                    int((len(self.scale) - 1) * value)
                                ],
                            ).colored,
                        )
                else:
                    self.buffer.put_char(
                        x, y, self.scale[int((len(self.scale) - 1) * value)]
                    )
        if self.info:
            self.buffer.put_at(
                0, 0, f"COLORS: {' '.join([str(val) for val in self.colors])}"
            )
            for i in range(1, 5):
                self.buffer.put_at(0, i, f"VAL {i}: {str(self.vals[i-1]):>3s} ")

    def func(self, x, y, a, b, n):
        """
        Helper function to calculate the plasma value given the four plasma values
        """
        return math.sin(
            math.sqrt(
                (x - self.buffer.width() * a) ** 2
                + 4 * ((y - self.buffer.height() * b)) ** 2
            )
            * math.pi
            / n
        )



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import random

from bruhcolor import bruhcolored
from ..bruhutil import LIFE_COLORS
from .noise_effect import NoiseEffect

class StarEffect(NoiseEffect):
    """
    Class for rendering out a blinking star effect. This is just a Noise effect with a predefined intensity.
    Ideally the background would be ' ' for the best effect, but the choice is yours.
    """

    def __init__(self, buffer, background, color_type="GREYSCALE"):
        super(StarEffect, self).__init__(buffer, background)

        self.stars = f"{background*(100 // self.background_length)}.*+"
        self.stars_length = len(self.stars)
        self.color_type = color_type

    def update_color_type(self, color_type):
        """
        Function to update the color of the stars
        :param color_type: color map
        """
        self.color_type = color_type

    def update_background(self, background):
        """
        Function to update the background of the efffect
        :param background: the new background
        """
        self.background = background
        self.background_length = len(background)
        self.stars = f"{background*(100 // self.background_length)}.*+"
        self.stars_length = len(self.stars)

    def render_frame(self, frame_number):
        """
        Function to update the next frame of the Stars effect
        """
        for y in range(self.buffer.height()):
            for x in range(self.buffer.width()):
                if random.random() < self.intensity:
                    self.buffer.put_char(
                        x,
                        y,
                        bruhcolored(
                            self.stars[random.randint(0, self.stars_length - 1)],
                            color=LIFE_COLORS[self.color_type][
                                random.randint(
                                    0, len(LIFE_COLORS[self.color_type]) - 1
                                )
                            ],
                        ).colored,
                    )



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import random

from ..bruhutil import WIND_DIRECTIONS
from .base_effect import BaseEffect


class RainEffect(BaseEffect):
    """
    Effect to emmulate the look of rain
    """

    def __init__(
        self,
        buffer,
        background,
        img_start_x=None,
        img_start_y=None,
        img_width=None,
        img_height=None,
        collision=False,
        intensity=1,
        swells=False,
        wind_direction="none",
    ):
        super(RainEffect, self).__init__(buffer, background)

        self.image_present = (
            True if img_start_x and img_start_y and img_width and img_height else False
        )
        self.collision = collision
        self.intensity = intensity
        self.swells = swells
        self.swell_direction = 1
        self.multiplier = 1
        self.wind_direction = (
            wind_direction if wind_direction in WIND_DIRECTIONS else "none"
        )
        self.wind_mappings = {
            "east": [".\\", -1, ["\\", "."]],
            "none": [".|", 0, ["|", "."]],
            "west": ["./", 1, ["/", "."]],
        }
        self._set_rain()

    def update_multiplier(self, val):
        """
        Update the multiplier value that relates to shift amount
        :param val: value to set the multiplier to
        """
        self.multiplier = val

    def update_wind_direction(self, direction):
        """
        Update the direction of the rain
        :param dircetion: direction for the rain to fall (east, west, none)
        """
        if direction in WIND_DIRECTIONS:
            self.wind_direction = direction
            self._set_rain()

    def _set_rain(self):
        """
        Function set the rain based on the intensity and wind direction
        """
        self.rain = f"{' ' * (1000 - self.intensity)}"
        if self.intensity > 50:
            self.rain += "."
        if self.intensity > 250:
            self.rain += "."
        if self.intensity > 500:
            self.rain += self.wind_mappings[self.wind_direction][0]
        self.rain_length = len(self.rain)

    def update_intensity(self, intensity):
        """
        Function to update the intensity of the rain
        :param intensity: intentisy value
        """
        if self.swells:
            if self.intensity == 900:
                self.swell_direction = -1
            if self.intensity == 0:
                self.swell_direction = 1
            self.intensity += self.swell_direction
        else:
            self.intensity = intensity if intensity < 1000 else 999
        self._set_rain()

    def update_collision(
        self,
        img_start_x,
        img_start_y,
        img_width,
        img_height,
        collision,
        smart_transparent=False,
        image_buffer=None,
    ):
        """
        Function to set whether or not to visually see the rain collide with the ground
        or images if they are present
        :param img_start_x: where the image starts on the screen
        :param img_start_y: where the image starts on the screen
        :param img_width:   the width of the image
        :param img_height:  the height of the image
        :param collision:   update collision variable
        :param smart_transparent: update smart_transparent
        :param image_buffer: the buffer that contains the image
        """
        self.image_present = (
            True if img_start_x and img_start_y and img_width and img_height else False
        )
        self.collision = collision
        if self.image_present:
            self.img_start_x = img_start_x
            self.img_start_y = img_start_y
            self.img_height = img_height
            self.img_width = img_width
            self.smart_transparent = smart_transparent
            self.image_buffer = image_buffer
        else:
            self.image_buffer = None

    def update_swells(self, swells):
        """
        Function to set whether the intensity should evolve on it's own
        :param swells: True / False
        """
        self.swells = swells

    def render_frame(self, frame_number):
        """
        Function to render the next frame of the Rain Effect
        """
        if self.swells:
            self.update_intensity(None)
        if frame_number == 0:
            self.buffer.put_at(
                0,
                0,
                "".join(
                    [
                        self.rain[random.randint(0, self.rain_length - 1)]
                        for _ in range(self.buffer.width())
                    ]
                ),
            )
        else:
            self.buffer.shift(
                self.wind_mappings[self.wind_direction][1] * self.multiplier
            )
            self.buffer.scroll(-1 * self.multiplier)
            self.buffer.put_at(
                0,
                0,
                "".join(
                    [
                        self.rain[random.randint(0, self.rain_length - 1)]
                        for _ in range(self.buffer.width())
                    ]
                ),
            )

            if self.collision:
                for y in range(self.buffer.height()):
                    for x in range(self.buffer.width()):
                        # Wipe prior frames impact
                        if self.buffer.get_char(x, y) == "v":
                            self.buffer.put_char(x, y, " ")
                        else:
                            if self.image_present:
                                # if we are inscope of the image we need to process impacts
                                if self.image_buffer:
                                    if 0 <= y + 1 < self.buffer.height():
                                        if (
                                            not self.image_buffer.buffer[y + 1][x]
                                            in [" ", None]
                                            and self.buffer.get_char(x, y)
                                            in self.wind_mappings[self.wind_direction][
                                                2
                                            ]
                                        ):
                                            self.buffer.put_char(x, y, "v")

                            # impacting the bottom
                            if y == self.buffer.height() - 1:
                                if (
                                    self.buffer.get_char(x, y)
                                    in self.wind_mappings[self.wind_direction][2]
                                ):
                                    self.buffer.put_char(x, y, "v")



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import random

from bruhcolor import bruhcolored
from .base_effect import BaseEffect


class NoiseEffect(BaseEffect):
    """
    Class for generating noise.
    :param intensity: randomness for the noise, higher the value the slower the effect (due to computation).
                      Will be a value 1 - 999
    :param color: whether or not ro color the noise
    """

    def __init__(self, buffer, background, intensity=200, color=False):
        super(NoiseEffect, self).__init__(buffer, background)

        self.intensity = (
            intensity / 1000 if intensity and 1 <= intensity <= 999 else 200 / 1000
        )

        self.noise = " !@#$%^&*()_+1234567890-=~`qazwsxedcrfvtgbyhnujmik,ol.p;/[']\QAZXSWEDCVFRTGBNHYUJM<KIOL>?:P{\"}|"
        self.noise_length = len(self.noise)
        self.color = color

    def update_intensity(self, intensity):
        """
        Function to update the intensity of the effect
        :param intensity: new intensity
        """
        self.intensity = (
            intensity / 1000 if intensity and 1 <= intensity <= 999 else 200 / 1000
        )

    def update_color(self, color, characters):
        """
        Function to enable / disable color for the effect
        :param color: True / False
        :param character: True / False to make characters visable
        """
        self.color = color
        self.characters = characters

    def render_frame(self, frame_number):
        """
        Function to render the next frame of the Noise effect
        """
        if self.color:
            for y in range(self.buffer.height()):
                for x in range(self.buffer.width()):
                    if random.random() < self.intensity:
                        if self.characters:
                            self.buffer.put_char(
                                x,
                                y,
                                bruhcolored(
                                    self.noise[
                                        random.randint(0, self.noise_length - 1)
                                    ],
                                    on_color=random.randint(0, 255),
                                ).colored,
                            )
                        else:
                            self.buffer.put_char(
                                x,
                                y,
                                bruhcolored(
                                    " ", on_color=random.randint(0, 255)
                                ).colored,
                            )
        else:
            for y in range(self.buffer.height()):
                for x in range(self.buffer.width()):
                    if random.random() < self.intensity:
                        self.buffer.put_char(
                            x, y, self.noise[random.randint(0, self.noise_length - 1)]
                        )



"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import random

from bruhcolor import bruhcolored
from ..bruhutil import TWINKLE_COLORS
from .base_effect import BaseEffect


class TWINKLE_SPEC:
    def __init__(self, char, value):
        self.char = char
        self.value = value
        self.fade = bruhcolored(self.char, TWINKLE_COLORS[self.value])
        self.mode = random.choice([1, -1])

    def __str__(self):
        return self.fade.colored

    def __repr__(self):
        return self.fade.colored

    def __len__(self):
        return 1

    def next(self):
        if self.value >= 23:
            self.mode = -1
        elif self.value <= 0:
            self.mode = 1

        self.value = self.value + self.mode

        self.fade = bruhcolored(self.char, TWINKLE_COLORS[self.value])
        return self

    def copy(self):
        new_TWINKLE_SPEC = TWINKLE_SPEC(self.char, self.value)
        new_TWINKLE_SPEC.mode = self.mode
        return new_TWINKLE_SPEC


class TwinkleEffect(BaseEffect):
    def __init__(self, buffer, background):
        super(TwinkleEffect, self).__init__(buffer, background)
        self.specs = []

    def render_frame(self, frame_number):
        if frame_number == 0:
            for y in range(self.buffer.height()):
                for x in range(self.buffer.width()):
                    if random.random() < 0.05:
                        new_TWINKLE_SPEC = TWINKLE_SPEC(".", random.randint(0, 23))
                        self.buffer.put_char(x, y, new_TWINKLE_SPEC)
                        self.specs.append((x, y))
        else:
            for x, y in self.specs:
                spec = self.buffer.get_char(x, y)
                self.buffer.put_char(x, y, spec.next().copy())
                del spec



class ScreenResizedError(Exception):
    """Exception raised when the screen is resized."""

    def __init__(self, message="The screen has been resized"):
        self.message = message
        super().__init__(self.message)

class InvalidEffectTypeError(Exception):
    """Exception raised when an invalid effect type is used."""

    def ____init__(self, message="Invalid effect type"):
        self.message = message
        super().__init__(self.message)

class InvalidImageError(Exception):
    """Exception raised when an invalid effect type is used."""

    def ____init__(self, message="Invalid image"):
        self.message = message
        super().__init__(self.message)

class InvalidPanRendererDirectionError(Exception):
    """Exception raised when an invalid effect type is used."""

    def ____init__(self, message="Invalid direction"):
        self.message = message
        super().__init__(self.message)



import sys


if sys.platform == "win32":
    from .bruheffect import AudioEffect
    from .demos import audio_demo


from .bruheffect import (
    BaseEffect,
    StaticEffect,
    OffsetEffect,
    NoiseEffect,
    StarEffect,
    PlasmaEffect,
    GameOfLifeEffect,
    RainEffect,
    MatrixEffect,
    ChatbotEffect,
    GradientNoise,
    Loading,
    StringStreamer,
    Key,
    Line,
    DrawLinesEffect,
    SnowEffect,
    TwinkleEffect,
    TWINKLE_SPEC,
)


from .bruhutil import (
    Screen,
    Buffer,
    bruhimage,
    get_image,
    text_to_image,
    LIFE_COLORS,
    LIFE_SCALES,
    PLASMA_COLORS,
    PLASMA_VALUES,
    GRADIENTS,
    VALID_DIRECTIONS,
    OLD_GREY_SCALES,
    GREY_SCALES,
    WIND_DIRECTIONS,
    NOISE,
    SNOWFLAKE_COLORS,
    FLAKE_WEIGHT_CHARS,
    TWINKLE_COLORS,
    VALID_INTERFACES,
    SNOWFLAKE_TYPES
)

from .bruhrenderer import (
    BaseRenderer,
    EffectRenderer,
    CenterRenderer,
    PanRenderer,
    FocusRenderer,
    BackgroundColorRenderer,
)

from .demos import (
    line_demo,
    plasma_demo,
    snow_demo,
    holiday,
    stars_demo,
    twinkle_demo,
    noise_demo,
    matrix_demo,
    gol_demo,
    rain_demo,
    offset_demo,
    static_demo,
    chatbot_demo,
)


__version__ = "0.2.69"
__valid_demos__ = [
    "static_demo",
    "offset_demo",
    "matrix_demo",
    "gol_demo",
    "rain_demo",
    "chatbot_demo" "line_demo",
    "plasma_demo",
    "snow_demo",
    "stars_demo",
    "twinkle_demo",
    "noise_demo" "holiday",
]

if sys.platform == "win32":
    __valid_demos__.append("audio_demo")


if sys.platform == "win32":
    __all__ = [
        "Screen",
        "plasma_demo",
        "line_demo",
        "holiday",
        "snow_demo",
        "stars_demo",
        "twinkle_demo",
        "noise_demo",
        "matrix_demo",
        "gol_demo",
        "rain_demo",
        "offset_demo",
        "static_demo",
        "audio_demo",
        "chatbot_demo",
        "AudioEffect",
        "BaseEffect",
        "StaticEffect",
        "OffsetEffect",
        "NoiseEffect",
        "StarEffect",
        "PlasmaEffect",
        "GameOfLifeEffect",
        "RainEffect",
        "MatrixEffect",
        "ChatbotEffect",
        "StringStreamer",
        "GradientNoise",
        "Loading",
        "Key",
        "DrawLinesEffect",
        "Line",
        "FLAKE",
        "SnowEffect",
        "Buffer",
        "BaseRenderer",
        "EffectRenderer",
        "CenterRenderer",
        "PanRenderer",
        "FocusRenderer",
        "BackgroundColorRenderer",
        "bruhimage",
        "__version__",
        "__valid_demos__",
        "LIFE_COLORS",
        "LIFE_SCALES",
        "PLASMA_COLORS",
        "PLASMA_VALUES",
        "GRADIENTS",
        "VALID_DIRECTIONS",
        "OLD_GREY_SCALES",
        "GREY_SCALES",
        "WIND_DIRECTIONS",
        "NOISE",
        "SNOWFLAKE_COLORS",
        "SNOWFLAKE_TYPES",
        "FLAKE_WEIGHT_CHARS",
        "TWINKLE_COLORS",
        "VALID_INTERFACES",
        "get_image",
        "text_to_image",
        "TwinkleEffect",
        "TWINKLE_SPEC",
    ]
else:
    __all__ = [
        "Screen",
        "plasma_demo",
        "line_demo",
        "holiday",
        "snow_demo",
        "stars_demo",
        "twinkle_demo",
        "noise_demo",
        "matrix_demo",
        "gol_demo",
        "rain_demo",
        "offset_demo",
        "static_demo",
        "chatbot_demo",
        "BaseEffect",
        "StaticEffect",
        "OffsetEffect",
        "NoiseEffect",
        "StarEffect",
        "PlasmaEffect",
        "GameOfLifeEffect",
        "RainEffect",
        "MatrixEffect",
        "ChatbotEffect",
        "StringStreamer",
        "GradientNoise",
        "Loading",
        "Key",
        "DrawLinesEffect",
        "Line",
        "FLAKE",
        "SnowEffect",
        "Buffer",
        "BaseRenderer",
        "EffectRenderer",
        "CenterRenderer",
        "PanRenderer",
        "FocusRenderer",
        "BackgroundColorRenderer",
        "images",
        "__version__",
        "__valid_demos__",
        "LIFE_COLORS",
        "LIFE_SCALES",
        "PLASMA_COLORS",
        "PLASMA_VALUES",
        "GRADIENTS",
        "VALID_DIRECTIONS",
        "OLD_GREY_SCALES",
        "GREY_SCALES",
        "WIND_DIRECTIONS",
        "NOISE",
        "FLAKES",
        "FLAKE_COLORS",
        "FLAKE_JUMPS",
        "NEXT_FLAKE_MOVE",
        "FLAKE_WEIGHT_CHARS",
        "FLAKE_FLIPS",
        "TWINKLE_COLORS",
        "VALID_INTERFACES",
        "get_image",
        "text_to_image",
        "TwinkleEffect",
        "TWINKLE_SPEC",
    ]




"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from ..bruhutil import VALID_DIRECTIONS
from .base_effect import BaseEffect


class OffsetEffect(BaseEffect):
    """
    Class for generating an offset-static backgorund.
    :new-param direction: which way the offset should go.
    """

    def __init__(self, buffer, background, direction="right"):
        super(OffsetEffect, self).__init__(buffer, background)
        self.direction = direction if direction in VALID_DIRECTIONS else "right"

    def update_direction(self, direction):
        """
        Function to update the direction of the offset
        :param direction: East / West
        """
        self.direction = direction if direction in VALID_DIRECTIONS else "right"
        self.buffer.clear_buffer()

    def render_frame(self, frame_number):
        """
        Function to render the next frame of the Offset effect
        """
        for y in range(self.buffer.height()):
            row = (
                self.background[y % self.background_length :]
                + self.background[: y % self.background_length]
            ) * (self.buffer.width() // self.background_length + self.background_length)
            if self.direction == "right":
                self.buffer.put_at(0, y, row[::-1])
            else:
                self.buffer.put_at(0, y, row)




"""
Copyright 2023 Ethan Christensen

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import sys
if sys.platform == "win32":
    import pyaudio
    import random

    import numpy as np

    from bruhcolor import bruhcolored
    from .base_effect import BaseEffect


    class AudioEffect(BaseEffect):
        def __init__(
            self, buffer, background: str, num_bands: int = 24, audio_halt: int = 10
        ):
            super(AudioEffect, self).__init__(buffer, background)
            self.FORMAT = pyaudio.paInt16
            self.CHANNELS = 1
            self.RATE = 44100
            self.CHUNK = 1024
            self.BANDS = num_bands + 1
            self.audio_halt = audio_halt
            self.bands = []
            self.p = pyaudio.PyAudio()
            self.upper_band_bound = self.buffer.height()
            self.band_ranges = self.generate_even_ranges(self.BANDS, 0, self.buffer.width())
            self.colors = [random.randint(0, 255) for _ in range(self.BANDS)]
            self.stream = self.p.open(
                format=self.FORMAT,
                channels=self.CHANNELS,
                rate=self.RATE,
                input=True,
                frames_per_buffer=self.CHUNK,
                stream_callback=self.process_audio,
            )
            self.gradient_mode = "extend"
            self.base_gradient = [
                232,
                233,
                235,
                237,
                239,
                241,
                243,
                245,
                247,
                249,
                251,
                253,
                255,
            ]
            self.gradient = []
            for _ in range(self.BANDS):
                self.gradient += self.base_gradient
            while len(self.gradient) < self.buffer.width():
                self.gradient += self.base_gradient
            self.true_gradient = [self.gradient[idx] for idx in range(self.buffer.width())]
            self.use_gradient = True
            self.non_gradient_color = 27
            self.orientation = "top"
            self.subtract_y = self.buffer.height()

        def set_audio_properties(
            self, num_bands=24, audio_halt=10, use_gradient=True, non_gradient_color=27
        ):
            self.BANDS = num_bands
            self.audio_halt = audio_halt
            self.band_ranges = self.generate_even_ranges(self.BANDS, 0, self.buffer.width())
            self.colors = [random.randint(0, 255) for _ in range(self.BANDS)]
            self.use_gradient = use_gradient
            self.non_gradient_color = non_gradient_color

        def evenly_distribute_original_values(self, original_list, desired_width):
            repeat_count = desired_width // len(original_list)
            extra_elements = desired_width % len(original_list)
            expanded_list = []
            for value in original_list:
                expanded_list.extend([value] * repeat_count)
                if extra_elements > 0:
                    expanded_list.append(value)
                    extra_elements -= 1
            return expanded_list

        def set_orientation(self, orientation):
            if orientation in ["top", "bottom"]:
                self.orientation = orientation
                if self.orientation == "bottom":
                    self.subtract_y = self.buffer.height()
                else:
                    self.subtract_y = 0

        def set_audio_gradient(
            self,
            gradient=[232, 233, 235, 237, 239, 241, 243, 245, 247, 249, 251, 253, 255],
            mode="extend",
        ):
            self.base_gradient = gradient
            self.gradient_mode = mode
            if self.gradient_mode == "repeat":
                self.gradient = []
                for _ in range(self.BANDS):
                    self.gradient += self.base_gradient
                while len(self.gradient) < self.buffer.width():
                    self.gradient += self.base_gradient
                self.true_gradient = [
                    self.gradient[idx] for idx in range(self.buffer.width())
                ]
            else:
                self.gradient = self.evenly_distribute_original_values(
                    gradient, self.buffer.width()
                )
                self.true_gradient = [
                    self.gradient[idx] for idx in range(self.buffer.width())
                ]

        def process_audio(self, data, frame_count, time_info, status):
            audio_array = np.frombuffer(data, dtype=np.int16)
            fft_result = np.fft.rfft(audio_array)
            magnitudes = np.abs(fft_result)
            band_width = len(magnitudes) // self.BANDS
            self.bands = [
                np.mean(magnitudes[i * band_width : (i + 1) * band_width])
                for i in range(self.BANDS)
            ]
            return (data, pyaudio.paContinue)

        def map_bands_to_range(self, N):
            min_band = min(self.bands)
            max_band = max(self.bands)
            rand_band = max_band - min_band if max_band != min_band else 1
            normalized_bands = [(band - min_band) / rand_band for band in self.bands]
            scaled_bands = [int(band * N) for band in normalized_bands]
            return scaled_bands

        def generate_even_ranges(self, groups, start, end):
            approximate_group_size = round((end - start) / groups)
            intervals = []
            for i in range(groups):
                group_start = start + i * approximate_group_size
                group_end = group_start + approximate_group_size
                intervals.append((group_start, min(group_end, end)))
            return intervals

        def render_frame(self, frame_number):
            if frame_number == 0:
                self.stream.start_stream()

            if frame_number % self.audio_halt == 0:
                try:
                    self.buffer.clear_buffer()
                    mapped_bands = self.map_bands_to_range(self.upper_band_bound)[1:]
                    for idx, band_group in enumerate(zip(mapped_bands, self.band_ranges)):
                        band, band_range = band_group
                        for y_change in range(band):
                            for x_change in range(*band_range):
                                if self.use_gradient:
                                    self.buffer.put_char(
                                        x_change,
                                        abs(self.subtract_y - y_change),
                                        bruhcolored(
                                            " ", on_color=self.true_gradient[x_change]
                                        ).colored,
                                        False,
                                    )
                                else:
                                    self.buffer.put_char(
                                        x_change,
                                        abs(self.subtract_y - y_change),
                                        bruhcolored(
                                            " ", on_color=self.non_gradient_color
                                        ).colored,
                                        False,
                                    )
                except:
                    pass



import sys
from .base_effect import BaseEffect
from .static_effect import StaticEffect
from .star_effect import StarEffect
from .chatbot_effect import ChatbotEffect, GradientNoise, Loading, StringStreamer, Key
from .draw_lines_effect import DrawLinesEffect, Line
from .game_of_life_effect import GameOfLifeEffect
from .matrix_effect import MatrixEffect
from .plasma_effect import PlasmaEffect
from .snow_effect import SnowEffect
from .twinkle_effect import TwinkleEffect, TWINKLE_SPEC
from .offset_effect import OffsetEffect
from .rain_effect import RainEffect
from .noise_effect import NoiseEffect


if sys.platform == "win32":
    from .audio_effect import AudioEffect
    __all__ = [
        "BaseEffect",
        "AudioEffect",
        "StaticEffect",
        "StarEffect",
        "ChatbotEffect",
        "DrawLinesEffect",
        "GameOfLifeEffect",
        "MatrixEffect",
        "PlasmaEffect",
        "SnowEffect",
        "TwinkleEffect",
        "OffsetEffect",
        "RainEffect",
        "NoiseEffect",
        "GradientNoise",
        "Loading",
        "StringStreamer",
        "Key",
        "Line",
        "TWINKLE_SPEC",
    ]
else:
    __all__ = [
        "BaseEffect",
        "StaticEffect",
        "StarEffect",
        "ChatbotEffect",
        "DrawLinesEffect",
        "GameOfLifeEffect",
        "MatrixEffect",
        "PlasmaEffect",
        "SnowEffect",
        "TwinkleEffect",
        "OffsetEffect",
        "RainEffect",
        "NoiseEffect",
        "GradientNoise",
        "Loading",
        "StringStreamer",
        "Key",
        "Line",
        "TWINKLE_SPEC",
    ]



from .bruhscreen import Screen
from .bruhffer import Buffer
from .bruhimage import get_image, text_to_image
from .bruhtypes import EffectType, Font, Image, valid_effect_types
from .bruherrors import ScreenResizedError, InvalidEffectTypeError, InvalidImageError
from .utils import (
    INF,
    VERTICAL,
    HORIZONTAL,
    VALID_EFFECTS,
    SNOWFLAKE_COLORS,
    FLAKE_WEIGHT_CHARS,
    GRADIENTS,
    GREY_SCALES,
    LIFE_COLORS,
    LIFE_SCALES,
    NOISE,
    OLD_GREY_SCALES,
    PLASMA_COLORS,
    PLASMA_VALUES,
    TWINKLE_COLORS,
    VALID_DIRECTIONS,
    WIND_DIRECTIONS,
    VALID_INTERFACES,
    SNOWFLAKE_TYPES,
    sleep,
)

__all__ = [
    "Screen",
    "Buffer",
    "get_image",
    "text_to_image",
    "INF",
    "VERTICAL",
    "HORIZONTAL",
    "VALID_EFFECTS",
    "SNOWFLAKE_COLORS",
    "FLAKE_WEIGHT_CHARS",
    "SNOWFLAKE_TYPES",
    "GRADIENTS",
    "GREY_SCALES",
    "LIFE_COLORS",
    "LIFE_SCALES",
    "NOISE",
    "OLD_GREY_SCALES",
    "PLASMA_COLORS",
    "PLASMA_VALUES",
    "TWINKLE_COLORS",
    "VALID_DIRECTIONS",
    "WIND_DIRECTIONS",
    "VALID_INTERFACES",
    "sleep",
    "Effect",
    "Font",
    "Image",
    "valid_effects",
    "ScreenResizedError",
    "InvalidEffectTypeError",
    "InvalidImageError"
]

